"""
Simulated SLM.
"""

import numpy as np
from .slm import SLM

class SimulatedSLM(SLM):
    """
    A simulated SLM to emulate physical artifacts of actual SLMs.

    Attributes
    ----------
    source : dict
        For a :class:`SimulatedSLM()`, :attr:`source` stores ``"amplitude_sim"`` and ``"phase_sim"``,
        which are used to compute the SLM's simulated far-field.

        ``"amplitude_sim"`` : numpy.ndarray
            User-defined source amplitude (with the dimensions of :attr:`shape`) on the SLM.
        ``"phase_sim"`` : numpy.ndarray
            User-defined source phase (with the dimensions of :attr:`shape`) on the SLM.
    """

    def __init__(self, resolution, pitch_um=(8,8), source=None, **kwargs):
        r"""
        Initialize simulated slm.

        Parameters
        ----------
        resolution
            The width and height of the camera in ``(width, height)`` form.

            Important
            ~~~~~~~~~
            This is the opposite of the numpy ``(height, width)``
            convention stored in :attr:`shape`.
        pitch_um : (float, float)
            Pixel pitch in microns. Defaults to 8 micron square pixels.
        source : dict
            See :attr:`source`. Defaults to uniform illumination with a flat phase if ``None``.
        **kwargs
            See :meth:`.SLM.__init__` for permissible options.
        """
        super().__init__(resolution, pitch_um=pitch_um, settle_time_s=0, **kwargs)

        if source is None:
            self.source["amplitude_sim"] = np.ones_like(self.grid[0])
            self.source["phase_sim"] = np.zeros_like(self.grid[0])
        else:
            # assert np.all([source[kw].shape == self.shape for kw in source.keys()]
            # ), "The shape of the provided phase profile must match the SLM resolution!"
            self.source.update(source)

            # Handle case where `source` only has real values from experiment
            if "amplitude_sim" not in source.keys():
                self.source["amplitude_sim"] = self.source["amplitude"]
                self.source["phase_sim"] = -self.source["phase"]

        self.write(None)

    def _write_hw(self, phase):
        """Updates SLM.display to implement various physical artifacts of SLMs."""

        # FUTURE: apply physical effects directly to SLM.display

        return



# The first line of the docstring might read awkward alone,
# but it's meant to parallel the slms module description.
"""
The sensor arrays used to measure results.
Computer vision hardware is connected to python by a myriad of SDKs, often provided by hardware
vendors. However, these SDKs likewise have a myriad of function names and hardware-specific
quirks. Thus, cameras in :mod:`slmsuite` are integrated as subclasses of
the abstract class :class:`.Camera`, which requires subclasses to implement a number of methods
relevant for SLM feedback (see below).
These subclasses are effectively wrappers for the given SDK, but also include
quality-of-life features such as image transformations (flips, rotates) and useful common methods.
A number of SDKs are supported, including:

-  :class:`.AlliedVision`, for AlliedVision cameras, through :mod:`vimba`
   (`GitHub <https://github.com/alliedvision/VimbaPython>`__),
-  :class:`.Cheetah640`, for Xenics Cheetah cameras, through the
   `Xenics SDK <https://www.xenics.com/software/>`_,
-  :class:`.FLIR`, for Teledyne FLIR cameras, through the :mod:`PySpin`
   interface of the `Spinnaker SDK <https://www.flir.com/products/spinnaker-sdk/>`_,
-  :class:`.MMCore`, for the general microscope control package Micro-Manager,
   through :mod:`pymmcore` (`GitHub <https://github.com/micro-manager/pymmcore>`__), and
-  :class:`.ThorCam`, for Thorlabs scientific cameras, through :mod:`thorlabs_tsi_sdk`
   and the `ThorCam SDK <https://www.thorlabs.com/software_pages/ViewSoftwarePage.cfm?Code=ThorCam>`_.

Tip
~~~~~~~~
While the superclass :class:`.Camera` only requires a small number of features to
be implemented as class functions, any further control of a camera interface can be
accessed by using the given SDK object directly (usually the attribute :attr:`cam` of
the subclass) or writing additional functions into the subclass.
"""



"""
Hardware control for Xenics camera via the :mod:`Xeneth` interface.

Python wrapper for Xenith C++ SDK 2.7 from Xenics; see xeneth-sdk.chm
there for additional documentation.
"""
import time
from ctypes import *

import numpy as np

from .camera import Camera

CLKFREQ = 80e6

### Imaging Constants from XFilters.h ###

# pylint: disable=C0301
# fmt: off
#ErrorCodes
I_OK                    = 0        # Success.
I_DIRTY                 = 1        # Internal.
E_BUG                   = 10000    # Generic.
E_NOINIT                = 10001    # Camera was not successfully initialised.
E_LOGICLOADFAILED       = 10002    # Invalid logic file.
E_INTERFACE_ERROR       = 10003    # Command interface failure.
E_OUT_OF_RANGE          = 10004    # Provided value is incapable of being produced by the hardware.
E_NOT_SUPPORTED         = 10005    # Functionality not supported by this camera.
E_NOT_FOUND             = 10006    # File/Data not found.
E_FILTER_DONE           = 10007    # Filter has finished processing and will be removed.
E_NO_FRAME              = 10008    # A frame was requested by calling GetFrame but none was available.
E_SAVE_ERROR            = 10009    # Couldn't save to file.
E_MISMATCHED            = 10010    # Buffer size mismatch.
E_BUSY                  = 10011    # The API can not read a temperature because the camera is busy.
E_INVALID_HANDLE        = 10012    # An unknown handle was passed to the C API.
E_TIMEOUT               = 10013    # Operation timed out.
E_FRAMEGRABBER          = 10014    # Frame grabber error.
E_NO_CONVERSION         = 10015    # GetFrame could not convert the image data to the requested format.
E_FILTER_SKIP_FRAME     = 10016    # Filter indicates the frame should be skipped.
E_WRONG_VERSION         = 10017    # Version mismatch.
E_PACKET_ERROR          = 10018    # The requested frame cannot be provided because at least one packet has been lost.
E_WRONG_FORMAT          = 10019    # The emissivity map you tried to set should be a 16 bit greyscale PNG.
E_WRONG_SIZE            = 10020    # The emissivity map you tried to set has the wrong dimensions (width height).
E_CAPSTOP               = 10021    # Internal
E_OUT_OF_MEMORY         = 10022    # An allocation failed because the system ran out of memory.

#ColourMode
COLOURMODE_8            = 0    # Intensity only
COLOURMODE_16           = 1    # Alias
COLOURMODE_PROFILE      = 2    # Uses a colour profile bitmap. See #LoadColourProfile()
COLOURMODE_INVERT       = 256   # Set this flag if an inversion of the colour profile is desired. eg: #ColourMode_8 | #ColourMode_Invert

#BlitType
WINDOW          = 0    # Blit the contents of the last captured frame directly to a Windows client device context using a Window handle (HWND) */
DEVICECONTEXT   = 1    # Blit the contents of the last captured frame to a specified device context.
                       #        This can be any device context (HDC) like a memory DC paint DC or a handle to a DC associated with a Graphics-object (C#)   */

#FrameType
FT_UNKNOWN      = -1   # Unknown invalid frame type
FT_NATIVE       = 0    # The native frame type of this camera (can be FT_8..,FT_16..,FT32.. check GetFrameType())
FT_8_BPP_GRAY   = 1    # 8-bit greyscale
FT_16_BPP_GRAY  = 2    # 16-bit greyscale (default for most of the Xenics branded cameras)
FT_32_BPP_GRAY  = 3    # 32-bit greyscale
FT_32_BPP_RGBA  = 4    # 32-bit colour RGBA      [B,G,R,A] Available for output conversion.
FT_32_BPP_RGB   = 5    # 32-bit colour RGB       [B,G,R]   Available for output conversion.
FT_32_BPP_BGRA  = 6    # 32-bit colour BGRA      [R,G,B,A]
FT_32_BPP_BGR   = 7     # 32-bit colour BGR       [R,G,B]

#XEnumerationFlag
XEF_NETWORK         = 0x00000001   # Network
XEF_SERIAL          = 0x00000002   # Serial
XEF_CAMERALINK      = 0x00000004   # CameraLink
XEF_GIGEVISION      = 0x00000008   # GigEVision
XEF_COAXPRESS       = 0x00000010   # CoaXPress
XEF_USB             = 0x00000020   # USB
XEF_USB3VISION      = 0x00000040   # USB3Vision
XEF_GENCP           = 0x00000080   # CameraLink GenCP
XEF_ENABLEALL       = 0x0000FFFF   # Enable all protocols.
XEF_USECACHED       = 0x01000000   # Use cached devices on enumeration.
XEF_RELEASECACHE    = 0x02000000   # Release internally cached devices.

#XFilterMessage
XMSGINIT                = 0    # [API->Filter Event] Called when the filter is being installed  ( (!) calling thread context)         */
XMSGCLOSE               = 1    # [API->Filter Event] Called when the filter is being removed    ( (!) calling thread context)         */
XMSGFRAME               = 2    # [API->Filter Event] Called after every frame grab              ( (!) grabbing thread context)        */
XMSGGETNAME             = 3    # [App->Filter Event] Retrieve filter name: the filter should copy a friendly string to msgparm        */
XMSGGETVALUE            = 4    # [Obsolete]                                                                                           */
XMSGSAVE                = 5    # [Obsolete]                                                                                           */
XMSGGETSTATUS           = 6    # [API->Filter Event] Retrieves a general purpose status message from the image filter                 */
XMSGUPDATEVIEWPORT      = 7    # [API->Filter Event] Instructs an image correction filter to update it's view port
                               #                            This message is sent to a filter upon changing the window of interest or when
                               #                            flipping image horizontally or vertically                                        */
XMSGCANPROCEED          = 8    # Used by image filters in in interactive mode to indicate acceptable image conditions                 */
XMSGGETINFO             = 9    # [Internal]          Used to query filter 'registers'                                                 */
XMSGSELECT              = 10   # [Obsolete]                                                                                           */
XMSGPROCESSEDFRAME      = 11   # [API->Filter Event] Sent after other filters have done their processing. Do not modify the frame data
                               #                            in response to this event.                                                       */
XMSGTIMEOUT             = 13   # [API->Filter Event] A camera time-out event was generated                                             */
XMSGISBUSY              = 16   # [Thermography]      Is the temperature filter recalculating - Used to check if the thermal filter is
                               #                            still updating it's linearisation tables                                         */
XMSGSETTROI             = 17   # [Imaging/Thermo]    Set the adu/temperature span in percent (see #XMsgSetTROIParms)                 */
XMSGLOAD                = 18   # [Obsolete]                                                                                           */
XMSGUNLOAD              = 19   # [Obsolete]                                                                                           */
XMSGADUTOTEMP           = 12   # [Thermography]      Convert an ADU value to a temperature (see #XFltADUToTemperature)                 */
XMSGGETEN               = 14   # [Obsolete]          Get temperature correction parameters (see #XMsgGetRadiometricParms)              */
XMSGSETEN               = 15   # [Obsolete]          Set temperature correction parameters (see #XMsgGetRadiometricParms)              */
XMSGTEMPTOADU           = 20   # [Thermography]      Convert a temperature to an ADU value (see #XFltTemperatureToADU)                 */
XMSGGETTVALUE           = 21   # [Thermography]      Retrieve an emissivity corrected value from a coordinate                         */
XMSGGETRADIOMETRICPARMS = 22   # [Thermography]      Get temperature correction parameters (see #XMsgTempParms)                        */
XMSGSETRADIOMETRICPARMS = 23   # [Thermography]      Set temperature correction parameters (see #XMsgTempParms)                        */
XMSGSERIALISE           = 100  # [App->Filter event] Serialise internal parameter state (write xml structure) see #XFltSetParameter    */
XMSGDESERIALISE         = 101  # [App->Filter event] Deserialise parameter state (read xml structure) see #XFltSetParameter            */
XMSGGETPRIORITY         = 102  # [Filter Management] Write the current filter priority to the long * provided in v_pMsgParm           */
XMSGSETFILTERSTATE      = 104  # [Filter Management] Enable or disable an image filter temporarily by sending 0/1 in v_pMsgParm       */
XMSGISSERIALISEDIRTY    = 105  # [Internal]                                                                                           */
XMSGSTOREHANDLE         = 106  # [Internal]          Start tracking the module handle for plugin image filters                        */
XMSGUPDATETINT          = 107  # [API->Filter event] Integration time change notification                                             */
XMSGLINADUTOTEMP        = 109  # [Thermography]      Convert a Linearized ADU value to a temperature (see #XFltADUToTemperatureLin)    */
XMSGLINTEMPTOADU        = 110  # [Thermography]      Convert a temperature to a Linearized ADU value (see #XFltTemperatureToADULin)    */
XMSGUPDATESPAN          = 111  # [API->Filter event] Span change notification                                                         */
XMSGUPDATEPALETTE       = 112  # [API->Filter event] Colour profile change notification                                               */
XMSGFILTERQUEUED        = 113  # [API->Filter event] A filter is queued                                                               */
XMSGFILTERREMOVED       = 114  # [API->Filter event] A filter is removed                                                              */
XMSGDRAWOVERLAY         = 200  # [API->Filter event] Draw the RGBA frame overlay, v_pMsgParm is the pointer to the RGBA data
                               #                            structure                                                                        */
XMSGLINEARISEOUTPUT     = 201  # [Thermography]      When specifying a v_pMsgParm that is non zero, starts linearising adu output     */
XMSGSETEMIMAP           = 202  # [Thermography]      Streams the main emissivity map to the thermal filter (16 bit png, 65535 = 1.0)  */
XMSGSETEMIMAPUSER       = 203  # [Thermography]      Stream a user emissivity map to the thermal filter (16 bit png, 65535 = 1.0,
                                #                            0 values are replaced by the emissivity in the main map)                         */
XMSGGETEMIMAP           = 204  # [Thermography]      Stream out the combined emissivity map                                           */
XMSGCLREMIMAP           = 205  # [Thermography]      Clear emissivity map                                                             */
XMSGCLREMIMAPUSER       = 206  # [Thermography]      Clear emissivity map (user)                                                      */
XMSGPUSHRANGE           = 207  # [Thermography]      Push a new linearization range to the thermal filter                             */
XMSGTHMFILTERSTATE      = 208  # [Thermography]      Filter event indicating thermal filter queue/removal                             */
XMSGTHMADJUSTSET        = 209  # [Thermography]      Set global offset & gain adu adjustment (pre-thermal conversion)                 */
XMSGTHMADJUSTGET        = 210  # [Thermography]      (see #XMsgTempAdjustmentParms)                                                      */

XMSGLOG                 = 211  # [Plugin->API]       Fire a log event to the end user application\n
                               #                            Target filter id: 0xffffffff                                                */
XMSGGETDELTAT           = 212  # [Internal]                                                                                      */
XMSGGETTINTRANGE        = 213  # [Plugin->API]       Request the exposure time range
                               #                            Target filter id: 0xffffffff                                                 */
XMSGCORRECTIONDIRTY         = 214  # [Internal]          The onboard thermography parameters have changed                             */
XMSGHASRADIANCEINFO         = 215  # [Thermography]      Check if the radiance information is available. This is needed to for emissivity correction */
XMSGCORRECTIONDATACHANGED   = 216  # [Internal]          New correction data is loaded                             */
XMSGPOSTPROCESS             = 217  # [Internal]          A post processing step is introduced in the software correction filter */

XMSGZOOMLENSCONNECT     = 300  # [Zoom lens]         Connect to the zoom lens on the specified port.  */
XMSGZOOMLENSGETSTATE    = 301  # [Zoom lens]         Get the current zoom/focus state from the zoom lens filter.  */
XMSGZOOMLENSSETSTATE    = 302  # [Zoom lens]         Set the current zoom/focus state in the zoom lens filter.    */
XMSGZOOMLENSGETINFO     = 303  # [Zoom lens]         Get some descriptive information about the connected lens.   */

XMSGUSER                = 24200 # If you develop your own image filter plugins, please use this constant to offset your messages. */

#XStatusMessage
XSLOADLOGIC         = 1    # Passed when loading the camera's main logic file                           */
XSLOADVIDEOLOGIC    = 2    # Passed when loading the camera's video output firmware                     */
XSDATASTORAGE       = 3    # Passed when accessing persistent data on the camera                        */
XSCORRECTION        = 4    # Passed when uploading correction data to the camera                        */
XSSELFSTART         = 5    # Passed when a self starting camera is starting (instead of XSLoadLogic)    */
XSMESSAGE           = 6    # String event
                                #  This status message is used to relay critical errors and events originating
                                #  from within the API.
                                #  Cam|PropLimit|property=number - A filter notifies you your user interface should limit the value of 'property' to 'number'
                                #  Cam|TemperatureFilter|RangeUpdate       - The thermography filter uses this to notify you of a span update.
                                #  Cam|TemperatureFilter|Off               - The thermography filter suggests the application to dequeue the filter.
                                #  Cam|InterfaceUpdate           - Internal, do not handle, returning E_BUG here causes the API to stop unpacking 'abcd.package'.packages to %appdata%/xenics/interface
XSLOADGRABBER       = 7    # Passed when loading the framegrabber                                                                                */
XSDEVICEINFORMATION = 8    # Device information passed when connecting a device, ulP is the lower part of the address. When using 64-bit the higher part of the address is stored in ulT */

#XGetFrameFlags
XGF_BLOCKING    = 1    # In blocking-mode the method does not return immediately with the return codes #E_NO_FRAME / #I_OK.
                       #  Instead the method waits for a frame and only returns until a frame was captured, or a time-out period has elapsed.

XGF_NOCONVERSION= 2    # Prevents internal conversion to 8 bit, specifying this flag reduces computation time, but prevents #SaveData() and the #Blit() method from working.
XGF_FETCHPFF    = 4    # Retrieve the per frame footer with frame timing information. Call XCamera::GetFrameFooterLength() to determine the increase in frame size.
XGF_RFU_1       = 8
XGF_RFU_2       = 16
XGF_RFU_3       = 32

#XSaveDataFlags
XSD_FORCE16             = 1    # Forces 16-bit output independent of the current #ColourMode-setting (only possible for PNG's)
XSD_FORCE8              = 2    # Forces 8-bit output independent of the current #ColourMode
XSD_ALIGNLEFT           = 4    # Left aligns 16-bit output (#XSD_Force16 | #XSD_AlignLeft)
XSD_SAVETHERMALINFO     = 8    # Save thermal conversion structure (only available when saving 16-bit PNGs)
XSD_RFU_0               = 16   # Reserved
XSD_RFU_1               = 32
XSD_RFU_2               = 64
XSD_RFU_3               = 128

#XSaveSettingFlags
XSS_SAVECAMERAPROPS     = 1    # Define property sources to save settings from.
XSS_SAVEGRABBERPROPS    = 2    #
XSS_SAVEALLPROPS        = 4    # Also save properties marked 'No persist'.
XSS_SS_RFU_3            = 8    #

#XLoadSettingsFlags
XSS_IGNORENAIS        = 1    # Ignore properties which do not affect the image.
XSS_LS_RFU_1          = 2    #
XSS_LS_RFU_2          = 4    #
XSS_LS_RFU_3          = 8    #

#XLoadCalibrationFlags
XLC_STARTSOFTWARECORRECTION     = 1    # Starts the software correction filter after unpacking the calibration data
XLC_RFU_1                       = 2
XLC_RFU_2                       = 4
XLC_RFU_3                       = 8

#XPropType
XTYPE_NONE              = 0x00000000

XTYPE_BASE_MASK         = 0x000000ff   # Type mask
XTYPE_ATTR_MASK         = 0xffffff00   # Attribute mask
XTYPE_BASE_NUMBER       = 0x00000001   # A number (floating)
XTYPE_BASE_ENUM         = 0x00000002   # An enumerated type (a choice)
XTYPE_BASE_BOOL         = 0x00000004   # Boolean (true / false / 1 / 0)
XTYPE_BASE_BLOB         = 0x00000008   # Binary large object
XTYPE_BASE_STRING       = 0x00000010   # String
XTYPE_BASE_ACTION       = 0x00000020   # Action (button)
XTYPE_BASE_RFU1         = 0x00000040   # RFU
XTYPE_BASE_RFU2         = 0x00000080   # RFU

XTYPE_BASE_MINMAX       = 0x00002000   # The property accepts the strings 'min' and 'max' to set the best achievable extremities.
XTYPE_BASE_READONCE     = 0x00001000   # Property needs to be read at start-up only
XTYPE_BASE_NOPERSIST    = 0x00000800   # Property shouldn't be persisted (saved & restored)
XTYPE_BASE_NAI          = 0x00000400   # Property does not affect image intensity level ('Not Affecting Image')
XTYPE_BASE_RW           = 0x00000300   # Write and read back
XTYPE_BASE_WRITEABLE    = 0x00000200   # Writeable properties have this set in their high byte
XTYPE_BASE_READABLE     = 0x00000100   # Readable properties have this set in their high byte

XTYPE_NUMBER            = 0x00000201   # Write only number
XTYPE_ENUM              = 0x00000202   # Write only enumeration
XTYPE_BOOL              = 0x00000204   # Write only boolean
XTYPE_BLOB              = 0x00000208   # Write only binary large object
XTYPE_STRING            = 0x00000210   # Write only string
XTYPE_ACTION            = 0x00000220   # Action (button)

XTYPE_RO_NUMBER         = 0x00000101   # Read only number
XTYPE_RO_ENUM           = 0x00000102   # Read only enumeration
XTYPE_RO_BOOL           = 0x00000104   # Read only boolean
XTYPE_RO_BLOB           = 0x00000108   # Read only binary large object
XTYPE_RO_STRING         = 0x00000110   # Read only string

XTYPE_RW_NUMBER         = 0x00000301   # R/W number
XTYPE_RW_ENUM           = 0x00000302   # R/W enumeration
XTYPE_RW_BOOL           = 0x00000304   # R/W boolean
XTYPE_RW_BLOB           = 0x00000308   # R/W binary large object
XTYPE_RW_STRING         = 0x00000310   # R/W string

#XDirectories
XDIR_FILTERDATA         = 0x0000   # Filter data (%APPDATA%/XenICs/Data/&lt;sessionnumber&gt;)
XDIR_SCRIPTROOT         = 0x0001   # Script root (%APPDATA%/XenICs/Interface/&lt;PID-number&gt;)
XDIR_CALIBRATIONS       = 0x0002   # Calibration folder (%ProgramFiles%/Xeneth/Calibrations)
XDIR_INSTALLDIR         = 0x0003   # Installation folder (%CommonProgramFiles%/XenICs/Runtime)
XDIR_PLUGINS            = 0x0004   # Plugin folder (%CommonProgramFiles%/XenICs/Runtime/Plugins)
XDIR_CACHEPATH          = 0x0005   # Cache folder (%APPDATA%/XenICs/Cache)
XDIR_SDKRESOURCES       = 0x0006   # SDK resource folder (%CommonProgramFiles%/XenICs/Runtime/Resources)
XDIR_XENETH             = 0x0007   # Xeneth installation directory
XDIR_GRABBERSCRIPTROOT  = 0x0008   # Script root (%APPDATA%/XenICs/Interface/&lt;FrameGrabber&gt;)

#XDeviceStates
XDS_AVAILABLE = 0x0    # The device is available to establish a connection.
XDS_BUSY = 0x1         # The device is currently in use.
XDS_UNREACHABLE = 0x2  # The device was detected but is unreachable.
# pylint: enable=C0301
# fmt: on


class _XDeviceInformation(Structure):
    """Structure to be filled with camera data from the dll inteface."""

    _fields_ = [
        ("size", c_int),
        ("name", (c_char * 64)),
        ("transport", (c_char * 64)),
        ("url", (c_char * 256)),
        ("address", (c_char * 64)),
        ("serial", c_uint),
        ("pid", c_uint),
        ("state", c_uint),
    ]


class Cheetah640(Camera):
    """
    Xeneth's Cheetah640 camera.

    Attributes
    ----------
    xeneth : windll
        Object to talk with the Xeneth SDK.
    cam : ptr
        Object to talk with the desired camera.
    profile : {'triggered', 'free'}
        Current pre-configured operation mode.

        ``'triggered'`` - Captures on external trigger \n
        ``'free'`` - Continuous capture

    frame_size : int
        Size of the frame in bytes for use in grabbing data for the buffer.
    frame_buffer : c_ushort array
        Buffer used to grab data from the camera and sdk interface.
    last_capture : numpy.ndarray
        Previous capture by get_image()
    last_tag : numpy.uint16
        Frame tag of last capture
    last_process_time : float
        Time to process last frame
    filters : dict
        Dictionary with enabled filters. See Xeneth documentation for more information.
    """

    ### Camera Interface ###

    def __init__(
            self,
            virtual=False,
            temperature=None,
            pitch_um=(20,20),
            verbose=True,
            **kwargs
        ):
        """
        Initialize camera. Default ``profile`` is ``'free'``.

        Parameters
        ----------
        virtual : bool
            Whether or not the camera is virtual.
        temperature : float or None
            Temperature in degrees celcius to set on startup. ``None`` defaults to no cooling.
        pitch_um : (float, float) OR None
            Fill in extra information about the pixel pitch in ``(dx_um, dy_um)`` form
            to use additional calibrations. Defaults to 20 micron square pixels.
        verbose : bool
            Whether or not to print camera initialization information.
        **kwargs
            See :meth:`.Camera.__init__` for permissible options.

        Raises
        ------
        RuntimeError
            If the camera is not reachable.
        """
        # Load the SDK
        self.xeneth = windll.LoadLibrary("xeneth64.dll")

        # Enumerate devices
        dev_count = c_uint()
        _ = self.xeneth.XCD_EnumerateDevices(None, byref(dev_count), XEF_CAMERALINK)
        device = _XDeviceInformation()
        _ = self.xeneth.XCD_EnumerateDevices(
            byref(device), byref(dev_count), XEF_USECACHED
        )

        # Open the first avaliable camera
        if virtual:
            if verbose:
                print("Opening connection to virtual camera")
            self.cam = self.xeneth.XC_OpenCamera(b"soft://0", 0, 0)
        else:
            if device.name.decode():
                if verbose:
                    print("Opening connection to %s" % device.name.decode())
                self.cam = self.xeneth.XC_OpenCamera(device.url, 0, 0)
                name = device.name.decode()
            else:
                raise RuntimeError(
                    "Camera not reachable! Close Xeneth GUI or check connections."
                )

        # Check that the camera opened properly
        if self.xeneth.XC_IsInitialised(self.cam):
            if verbose:
                print("Camera initialized, loading properties ...")

            super().__init__(
                self.xeneth.XC_GetWidth(self.cam),
                self.xeneth.XC_GetHeight(self.cam),
                bitdepth=12,
                pitch_um=pitch_um,
                name=name,
                **kwargs
            )

            # Initialize a 16-bit buffer for a single frame; initialize to 0
            self.frame_size = self.xeneth.XC_Getframe_size(self.cam)
            self.frame_buffer = (c_ushort * int(self.frame_size / 2))(0)
            self.last_capture = np.empty(self.shape)
            self.last_tag = np.uint16(0)  # Frame tag of last capture
            self.last_process_time = 0  # Time to process last frame

            # Make a filter dictionary
            self.filters = {}

            # Setup base imaging configuration
            if not virtual:
                # Starts a circular buffer w/ 4k frames (anything larger crashes???)
                self.setup_grabber(frames=4000)
                # Max out the API buffer to
                self.set_buffer_api(frames=64)
                # Enable frame tags (frame number in first two pixels)
                self.enable_frametags(True)
                if temperature is None:
                    # Disable high fan speed for stability
                    self.enable_cooling(False)
                else:
                    self.enable_cooling()
                    # Set Peltier setpoint
                    self.set_temperature(temperature)
                    # Flip X, Flip Y readout by default (used in onboard cal)
                self.set_readout_orientation(True, True)
                # Enable output trigger to readout integration periods
                self.setup_output_trigger()
                # Set max frame rate to basically unlimited (exposure-limited)
                self.set_framerate(0)

                # Load visible setup by default
                self.setup("free")
        else:
            print("Initialization failed")

    def close(self):
        """See :meth:`.Camera.close`."""
        if self.xeneth.XC_IsInitialised(self.cam):
            print("Closing connection to camera...")
            self.xeneth.XC_CloseCamera(self.cam)
        else:
            print("Camera not open!")

    def reset(self):
        """See :meth:`.Camera.reset`."""
        self.close()
        self.__init__()

    ### Property Configuration ###

    def get_property_status(self, save_file_path=None, verbose=True):
        """
        List and get status of all addressable properties on camera.

        Parameters
        ----------
        save_file_path : str or None
            If not ``None``, the property status results will be saved to this path.
        verbose : True
            Prints property results if ``True``.
        """
        if self.xeneth.XC_IsInitialised(self.cam):
            if verbose:
                print("Getting camera properties...")
            property_count = self.xeneth.XC_GetPropertyCount(self.cam)
            property_name = (c_char * 128)(0)
            property_category = (c_char * 128)(0)
            property_range = (c_char * 128)(0)
            property_unit = (c_char * 128)(0)
            property_type = c_int()  # XType_None
            lvalue = c_long(0)
            fvalue = c_double(0)
            cvalue = c_char(0)

            # Iterate over each property and output details such as name, type, value
            for x in range(property_count):
                self.xeneth.XC_Getproperty_name(self.cam, x, property_name, 128)
                self.xeneth.XC_Getproperty_category(
                    self.cam, property_name, property_category, 128
                )
                self.xeneth.XC_Getproperty_type(
                    self.cam, property_name, byref(property_type)
                )
                self.xeneth.XC_Getproperty_range(
                    self.cam, property_name, property_range, 128
                )
                self.xeneth.XC_Getproperty_unit(
                    self.cam, property_name, property_unit, 128
                )

                if verbose:
                    print(
                        "Property[%d]    Category: %s"
                        % (x, property_category.value.decode())
                    )
                    print(
                        "Property[%d]        Name: %s"
                        % (x, property_name.value.decode())
                    )
                    print(
                        "Property[%d]       Flags: %s"
                        % (
                            x,
                            (
                                "MinMax | "
                                if property_type.value & XTYPE_BASE_MINMAX
                                else ""
                            )
                            + (
                                "ReadOnce | "
                                if property_type.value & XTYPE_BASE_READONCE
                                else ""
                            )
                            + (
                                "NoPersist | "
                                if property_type.value & XTYPE_BASE_NOPERSIST
                                else ""
                            )
                            + ("NAI | " if property_type.value & XTYPE_BASE_NAI else "")
                            + (
                                "Writeable | "
                                if property_type.value & XTYPE_BASE_WRITEABLE
                                else ""
                            )
                            + (
                                "Readable | "
                                if property_type.value & XTYPE_BASE_READABLE
                                else ""
                            ),
                        )
                    )

                # The following output depends on the property type.
                type_num = property_type.value & XTYPE_BASE_MASK

                # Write-only number
                if type_num == XTYPE_BASE_NUMBER:
                    # Check camera doc for float vs. long
                    self.xeneth.XC_GetPropertyValueL(
                        self.cam, property_name, byref(lvalue)
                    )
                    self.xeneth.XC_GetPropertyValueF(
                        self.cam, property_name, byref(fvalue)
                    )
                    if verbose:
                        print("Property[%d]        Type: Number" % x)
                        print(
                            "Property[%d]       Range: %s"
                            % (x, property_range.value.decode())
                        )
                        print(
                            "Property[%d]        Unit: %s"
                            % (x, property_unit.value.decode())
                        )
                        print("Property[%d]  Long value: %lu" % (x, lvalue.value))
                        print("Property[%d] Float value: %f" % (x, fvalue.value))

                # Write-only enumeration
                elif type_num == XTYPE_BASE_ENUM:
                    self.xeneth.XC_GetPropertyValueL(
                        self.cam, property_name, byref(lvalue)
                    )
                    if verbose:
                        print("Property[%d]        Type: Enum" % x)
                        print(
                            "Property[%d]       Range: %s"
                            % (x, property_range.value.decode())
                        )
                        print("Property[%d]       Value: %lu" % (x, lvalue.value))

                # Boolean TF
                elif type_num == XTYPE_BASE_BOOL:
                    self.xeneth.XC_GetPropertyValueL(
                        self.cam, property_name, byref(lvalue)
                    )
                    if verbose:
                        print("Property[%d]        Type: Bool" % x)
                        print(
                            "Property[%d]       Value: %s"
                            % (x, "True" if lvalue.value == 1 else "False")
                        )

                # Binary large object (BLOB)
                elif type_num == XTYPE_BASE_BLOB:
                    if verbose:
                        print("Skipping BLOB information collection to save time...")

                # String
                elif type_num == XTYPE_BASE_STRING:
                    # Get the string.
                    cvalue = (c_char * 128)(0)
                    self.xeneth.XC_GetPropertyValue(
                        self.cam, property_name, cvalue, 128
                    )
                    if verbose:
                        print("Property[%d]        Type: String" % x)
                        print("Property[%d]       Value: %s" % (x, cvalue.value))

                if verbose:
                    print("")

                # Save current settings to file if desired
                if save_file_path is not None:
                    fname = save_file_path + ".xcf"
                    print('Saving settings to file "{}"...'.format(fname))
                    self.xeneth.XC_SaveSettings(self.cam, fname)
        else:
            print("Camera not open!")

    def configure(self, format_file):
        """
        Loads pre-stored imaging profile from XC_SaveSettings XCF file

        Parameters
        ----------
        format_file : str
            Path to XCF file containing format information
        """
        if self.xeneth.XC_IsInitialised(self.cam):
            print("Loading settings from {}.xcf ....".format(format_file))
            self.xeneth.XC_LoadSettings(self.cam, format_file)
        else:
            print("Camera not open!")

    def get_exposure(self):
        """See :meth:`.Camera.get_exposure`."""
        exposure_old = c_double(0.0)
        err1 = self.xeneth.XC_GetPropertyValueF(
            self.cam, b"IntegrationTime", byref(exposure_old)
        )
        if err1:
            print("\nWarning -- error encountered! Error code: %d" % (err1))
        return exposure_old.value / 1e6

    def set_exposure(self, exposure_s, verbose=True):
        """
        See :meth:`.Camera.set_exposure`.

        Parameters
        ----------
        verbose : bool
            Whether or not to print extra information.
        """
        print(
            "Setting integration time to %1.3f ms...          " % (exposure_s * 1e3),
            end="\r",
        )
        exposure_old = c_double(0.0)
        exposure = c_double(exposure_s * 1e6)  # us
        err1 = self.xeneth.XC_GetPropertyValueF(
            self.cam, b"IntegrationTime", byref(exposure_old)
        )
        err2 = self.xeneth.XC_SetPropertyValueF(
            self.cam, b"IntegrationTime", exposure, ""
        )
        err3 = self.xeneth.XC_GetPropertyValueF(
            self.cam, b"IntegrationTime", byref(exposure)
        )
        if verbose:
            print(
                "\nPrevious integration time: %ld us\nNew integration time: %ld us"
                % (exposure_old.value, exposure.value)
            )
        if err1 or err2 or err3:
            print(
                "\nWarning -- error encountered! Error codes: %d, %d, %d"
                % (err1, err2, err3)
            )

    def set_framerate(self, framerate):
        """
        Set the camera framerate.

        Parameters
        ----------
        frame : int
            The framerate in fps.
        """
        rate = c_long(int(framerate))
        print("Setting frame rate to %d Hz..." % (rate.value))
        rate_old = c_long(0)
        err1 = False
        err1 = self.xeneth.XC_GetPropertyValueL(self.cam, b"FrameRate", byref(rate_old))
        err2 = self.xeneth.XC_SetPropertyValueL(self.cam, b"FrameRate", rate, "")
        err3 = self.xeneth.XC_GetPropertyValueL(self.cam, b"FrameRate", byref(rate))
        print(
            "Previous frame rate: %d Hz\nNew frame rate: %d Hz"
            % (rate_old.value, rate.value)
        )
        if err1 or err2 or err3:
            print(
                "Warning -- error encountered! Error codes: %d, %d, %d"
                % (err1, err2, err3)
            )

    def get_frame_footer_length(self):
        """
        Get the length of software frame tags.

        Returns
        -------
        int
        """
        return self.xeneth.XC_GetFrameFooterLength(self.cam)

    def set_buffer_api(self, frames=64):
        """
        Set the number of API-facing buffer frames.

        Parameters
        ----------
        frames : int
            Number of buffer frames to allocate.
        """
        print("Setting API buffer frame count to %d..." % (frames))
        frame_current = c_ulong(0)
        err1 = self.xeneth.XC_GetPropertyValueL(
            self.cam, b"_API_FPC_BFRNUM", byref(frame_current)
        )
        print("Previous API buffer size: %d frames" % (frame_current.value))
        err2 = self.xeneth.XC_SetPropertyValueL(
            self.cam, b"_API_FPC_BFRNUM", c_long(frames), ""
        )
        err3 = self.xeneth.XC_GetPropertyValueL(
            self.cam, b"_API_FPC_BFRNUM", byref(frame_current)
        )
        print("     New API buffer size: %d frames" % (frame_current.value))
        if err1 or err2 or err3:
            print(
                "Warning -- error encountered! Error codes: %d, %d, %d"
                % (err1, err2, err3)
            )

    def set_timeout_api(self, timeout_ms=10000):
        """
        Set the get_frame time allowed before issuing E_NOFRAME error.

        Warning
        ~~~~~~~~
        Implementation unfinished and untested.

        Parameters
        ----------
        timeout_ms : int
            Time in ms to wait for blocking frame capture.
        """
        print("Setting API timeout to %d ms..." % (timeout_ms))
        timeout_current = c_ulong(0)
        err1 = self.xeneth.XC_GetPropertyValueL(
            self.cam, b"_API_GETFRAME_TIMEOUT", byref(timeout_current)
        )
        print("Previous API timeout: %d ms" % (timeout_current.value))
        err2 = self.xeneth.XC_SetPropertyValueL(
            self.cam, b"_API_GETFRAME_TIMEOUT", c_long(timeout_ms), ""
        )
        err3 = self.xeneth.XC_GetPropertyValueL(
            self.cam, b"_API_GETFRAME_TIMEOUT", byref(timeout_current)
        )
        print("     New API timeout: %d ms" % (timeout_current.value))
        if err1 or err2 or err3:
            print(
                "Warning -- error encountered! Error codes: %d, %d, %d"
                % (err1, err2, err3)
            )

    def set_temperature(self, temp_c):
        """
        Set the camera temperature setpoint.

        Parameters
        ----------
        temp_c : float
            Temperature in degrees celcius to set on startup.
        """
        print("Setting settle temperature to %1.2fC..." % (temp_c))
        temp_current = c_double(0)
        err1 = self.xeneth.XC_GetPropertyValueF(
            self.cam, b"SettleTemperature", byref(temp_current)
        )
        print("Previous set temperature: %1.2f degC" % (temp_current.value - 273.15))
        err2 = self.xeneth.XC_SetPropertyValueF(
            self.cam, b"SettleTemperature", c_double(temp_c + 273.15), ""
        )
        err3 = self.xeneth.XC_GetPropertyValueF(
            self.cam, b"SettleTemperature", byref(temp_current)
        )
        print("     New set temperature: %1.2f degC" % (temp_current.value - 273.15))
        if err1 or err2 or err3:
            print(
                "Warning -- error encountered! Error codes: %d, %d, %d"
                % (err1, err2, err3)
            )

    def get_temperature(self):
        """
        Get the camera temperature setpoint.

        Returns
        -------
        float
            Temperature in degrees celcius to set on startup. -1 if the temperature
            could not be read.
        """
        temp = -1.0
        temp_current = c_double(0)
        err = self.xeneth.XC_GetPropertyValueF(
            self.cam, b"Temperature", byref(temp_current)
        )

        if err:
            print("Error while reading sensor temperature; code: " + str(err))
        else:
            temp = temp_current.value - 273.15

        return temp

    def set_readout_orientation(self, flip_x=True, flip_y=True):
        """
        Sets direction of pixel readout from focal plane.

        Parameters
        ----------
        flip_x, flip_y : tuple(bool)
            Whether a flip will be applied in the X or Y directions.
        """
        errs = []
        flip_x = c_long(int(flip_x))
        flip_y = c_long(int(flip_y))
        errs.append(
            self.xeneth.XC_SetPropertyValueL(self.cam, b"ReadoutFlipX", flip_x, "")
        )
        errs.append(
            self.xeneth.XC_SetPropertyValueL(self.cam, b"ReadoutFlipY", flip_y, "")
        )
        errs.append(
            self.xeneth.XC_GetPropertyValueL(self.cam, b"ReadoutFlipX", byref(flip_x))
        )
        errs.append(
            self.xeneth.XC_GetPropertyValueL(self.cam, b"ReadoutFlipX", byref(flip_x))
        )
        print(
            "Readout orientation set to (flip_x,flip_y) = (%s,%s)"
            % (str(bool(flip_x.value)), str(bool(flip_y.value)))
        )

        if any(errs):
            print("Warning! Errors detected in trigger setup. List: ", errs)

    def enable_frametags(self, enable=False):
        """
        Adds captured frame number to first two pixels of an image.
        Disabled by default to prevent issues with autoexposure.

        Parameters
        ----------
        enable : bool
            Whether to enable or disable.
        """
        if enable:
            err = self.xeneth.XC_SetPropertyValueE(self.cam, b"FrameMarker", b"Enabled")
        else:
            err = self.xeneth.XC_SetPropertyValueE(
                self.cam, b"FrameMarker", b"Disabled"
            )
        if err:
            print("Error setting frame tags! Code: " + str(err))
        else:
            print("Frame tags set...")

    def setup_input_trigger(
        self,
        mode=0,
        delay=0,
        source=0,
        skip=0,
        fpt=1,
        verbose=False,
    ):
        """
        Configure capture control via triggering.

        Parameters
        ----------
        mode : int
            Trigger properties.\n
            0 means free running.\n
            1 means level.\n
            2 means rising edge.\n
            3 means falling edge.
        delay : float
            Trigger delay in microseconds.
        source : int
            Source of the input trigger.\n
            0 means trigger in.\n
            1 means software.\n
            2 means CameraLink CC1.
        skip : int
            Number of frames to skip after trigger.
        fpt : int
            Frames per trigger.
        verbose : bool
            Enable debug printout.
        """
        trigger_modes = {
            0: b"Free running",
            1: b"Level",
            2: b"Rising edge",
            3: b"Falling edge",
        }
        trigger_sources = {0: b"Trigger in", 1: b"Software", 2: b"CameraLink CC1"}
        errs = []

        # Get current trigger setup
        mode_old = (c_char * 128)(0)
        errs.append(
            self.xeneth.XC_GetPropertyValueE(
                self.cam, b"TriggerMode", byref(mode_old), 128
            )
        )
        delay_old = c_double(0)
        errs.append(
            self.xeneth.XC_GetPropertyValueF(
                self.cam, b"TriggerInputDelay", byref(delay_old)
            )
        )
        source_old = (c_char * 128)(0)
        errs.append(
            self.xeneth.XC_GetPropertyValueE(
                self.cam, b"TriggerSource", byref(source_old), 128
            )
        )
        skip_old = c_long(0)
        errs.append(
            self.xeneth.XC_GetPropertyValueL(self.cam, b"TriggerSkip", byref(skip_old))
        )
        fpt_old = c_long(0)
        errs.append(
            self.xeneth.XC_GetPropertyValueL(self.cam, b"NrOfFrames", byref(fpt_old))
        )
        if verbose:
            print(
                "Original trigger setup: Mode - %s | Delay - %1.2fus "
                "| Source - %s | NSkip - %d | FPT - %d"
                % (
                    mode_old.value.decode(),
                    delay_old.value,
                    source_old.value.decode(),
                    skip_old.value,
                    fpt_old.value,
                )
            )

        # Set desired trigger setup
        errs.append(
            self.xeneth.XC_SetPropertyValueL(self.cam, b"NrOfFrames", c_long(fpt), "")
        )
        errs.append(
            self.xeneth.XC_SetPropertyValueE(
                self.cam, b"TriggerMode", trigger_modes[mode]
            )
        )
        errs.append(
            self.xeneth.XC_SetPropertyValueF(
                self.cam, b"TriggerInputDelay", c_double(delay), ""
            )
        )
        errs.append(
            self.xeneth.XC_SetPropertyValueE(
                self.cam, b"TriggerSource", trigger_sources[source]
            )
        )
        errs.append(
            self.xeneth.XC_SetPropertyValueL(self.cam, b"TriggerSkip", c_long(skip), "")
        )

        # Get final trigger setup
        errs.append(
            self.xeneth.XC_GetPropertyValueE(
                self.cam, b"TriggerMode", byref(mode_old), 128
            )
        )
        errs.append(
            self.xeneth.XC_GetPropertyValueF(
                self.cam, b"TriggerInputDelay", byref(delay_old)
            )
        )
        errs.append(
            self.xeneth.XC_GetPropertyValueE(
                self.cam, b"TriggerSource", byref(source_old), 128
            )
        )
        errs.append(
            self.xeneth.XC_GetPropertyValueL(self.cam, b"TriggerSkip", byref(skip_old))
        )
        errs.append(
            self.xeneth.XC_GetPropertyValueL(self.cam, b"NrOfFrames", byref(fpt_old))
        )
        if verbose:
            print(
                "     New trigger setup: Mode - %s | Delay - %1.2fus "
                "| Source - %s | NSkip - %d | FPT - %d"
                % (
                    mode_old.value.decode(),
                    delay_old.value,
                    source_old.value.decode(),
                    skip_old.value,
                    fpt_old.value,
                )
            )

        if any(errs):
            print("Warning! Errors detected in trigger setup. List: ", errs)

    def setup_output_trigger(
        self,
        enable=1,
        mode=1,
        source=2,
        delay=0,
        width=10,
        verbose=False,
    ):
        """
        Configures output trigger.

        Parameters
        ----------
        enable : bool
            Whether to enable the output trigger.
        mode : int
            Trigger properties.\n
            0 means active low.\n
            1 means active high.
        source : int
            Source of the input trigger.\n
            0 means integration start.\n
            1 means trigger input.\n
            2 means integration period.
        delay : float
            Trigger delay in microseconds.
        width : float
            Number of frames to skip after trigger.
        verbose : bool
            Enable debug printout.
        """

        trigger_status = {0: b"Off", 1: b"On"}
        trigger_modes = {0: b"Active low", 1: b"Active high"}
        trigger_sources = {
            0: b"Integration start",
            1: b"Trigger input",
            2: b"Integration period",
        }
        errs = []

        # Get current trigger setup
        status_old = (c_char * 128)(0)
        errs.append(
            self.xeneth.XC_GetPropertyValueE(
                self.cam, b"TriggerOutEnable", byref(status_old), 128
            )
        )
        mode_old = (c_char * 128)(0)
        errs.append(
            self.xeneth.XC_GetPropertyValueE(
                self.cam, b"TriggerOutMode", byref(mode_old), 128
            )
        )
        delay_old = c_double(0)
        errs.append(
            self.xeneth.XC_GetPropertyValueF(
                self.cam, b"TriggerOutDelay", byref(delay_old)
            )
        )
        source_old = (c_char * 128)(0)
        errs.append(
            self.xeneth.XC_GetPropertyValueE(
                self.cam, b"TriggerOutSource", byref(source_old), 128
            )
        )
        width_old = c_long(0)
        errs.append(
            self.xeneth.XC_GetPropertyValueL(
                self.cam, b"TriggerOutWidth", byref(width_old)
            )
        )
        if verbose:
            print(
                "Original trigger setup: Status - %s | Mode - %s | Delay - %1.2fus "
                "| Source - %s | Width - %d us"
                % (
                    status_old.value.decode(),
                    mode_old.value.decode(),
                    delay_old.value,
                    source_old.value.decode(),
                    width_old.value,
                )
            )

        # Set desired trigger setup
        errs.append(
            self.xeneth.XC_SetPropertyValueE(
                self.cam, b"TriggerOutEnable", trigger_status[enable]
            )
        )
        errs.append(
            self.xeneth.XC_SetPropertyValueE(
                self.cam, b"TriggerOutMode", trigger_modes[mode]
            )
        )
        errs.append(
            self.xeneth.XC_SetPropertyValueF(
                self.cam, b"TriggerOutDelay", c_double(delay), ""
            )
        )
        errs.append(
            self.xeneth.XC_SetPropertyValueE(
                self.cam, b"TriggerOutSource", trigger_sources[source]
            )
        )
        errs.append(
            self.xeneth.XC_SetPropertyValueF(
                self.cam, b"TriggerOutWidth", c_double(width), ""
            )
        )

        # Get final trigger setup
        errs.append(
            self.xeneth.XC_GetPropertyValueE(
                self.cam, b"TriggerOutEnable", byref(status_old), 128
            )
        )
        errs.append(
            self.xeneth.XC_GetPropertyValueE(
                self.cam, b"TriggerOutMode", byref(mode_old), 128
            )
        )
        errs.append(
            self.xeneth.XC_GetPropertyValueF(
                self.cam, b"TriggerOutDelay", byref(delay_old)
            )
        )
        errs.append(
            self.xeneth.XC_GetPropertyValueE(
                self.cam, b"TriggerOutSource", byref(source_old), 128
            )
        )
        errs.append(
            self.xeneth.XC_GetPropertyValueL(
                self.cam, b"TriggerOutWidth", byref(width_old)
            )
        )
        if verbose:
            print(
                "     New trigger setup: Status - %s | Mode - %s | Delay - %1.2fus "
                "| Source - %s | Width - %d us"
                % (
                    status_old.value.decode(),
                    mode_old.value.decode(),
                    delay_old.value,
                    source_old.value.decode(),
                    width_old.value,
                )
            )

        if any(errs):
            print("Warning! Errors detected in trigger setup. List: ", errs)

    def setup_grabber(self, mode=0, frames=4000):
        """
        Setup frame grabber capture mode.

        Parameters
        ----------
        mode : int
            Capture mode.\n
            0 means circular buffer.\n
            1 means waits till all frames grabbed before restarting capture.\n
            2 means non-circular (Stops grabbing when buffer filled).
        frames : int
            Number of frames to capture.
        """
        errs = []
        modes = {
            0: b"Preview",  # Circular buffer
            1: b"Synchronous bursts",  # Waits till all frames grabbed before restarting capture
            2: b"Synchronous burst",  # Non-circular (Stops grabbing when buffer filled)
        }
        mode_old = (c_char * 128)(0)
        errs.append(
            self.xeneth.XC_GetPropertyValueE(self.cam, b"Mode", byref(mode_old), 128)
        )
        print("Previous capture mode: %s" % mode_old.value.decode())
        errs.append(self.xeneth.XC_SetPropertyValueE(self.cam, b"Mode", modes[mode]))
        errs.append(
            self.xeneth.XC_GetPropertyValueE(self.cam, b"Mode", byref(mode_old), 128)
        )
        print("     New capture mode: %s" % mode_old.value.decode())

        # Set the buffer size
        errs.append(
            self.xeneth.XC_SetPropertyValueL(
                self.cam, b"FrameCount", c_long(frames), ""
            )
        )
        frames = c_long(0)
        errs.append(
            self.xeneth.XC_GetPropertyValueL(
                self.cam, b"FrameCount", byref(frames), 128
            )
        )
        print("Buffer frame count set to %d frames" % frames.value)

        if any(errs):
            print("Warning! Error(s) encountered: ", errs)

    def set_woi(self, woi=None, verbose=False):
        """See :meth:`.Camera.set_woi`

        Parameters
        ----------
        verbose : bool
            Enable debug printout.
        """
        if woi is None:
            woi = (0, self.default_shape[1], 0, self.default_shape[0])

        woi = (woi[0], woi[1] - woi[0], woi[2], woi[3] - woi[2])

        # If collecting, stop
        if self.isCapturing():
            self.stopCapture()

        # Get current WOI setup
        report_str = "Original WOI setup: "
        woi_prop = [b"WoiSX(0)", b"WoiEX(0)", b"WoiSY(0)", b"WoiEY(0)"]
        prop_val = c_long(0)
        errs = []
        for prop in woi_prop:
            errs.append(
                self.xeneth.XC_GetPropertyValueL(self.cam, prop, byref(prop_val))
            )
            report_str += "%s: %d | " % (prop.decode(), prop_val.value)
        if verbose:
            print(report_str)

        # Conservatively round inputs (make wider than requested) based on cam. reqts.
        min_w_factor = 16
        min_h_factor = 4
        if (woi[0]) % min_w_factor:
            woi[0] = max([woi[0] - woi[0] % min_w_factor, 0])
        if (woi[1] - woi[0] + 1) % min_w_factor:
            woi[1] = (
                woi[1] + min_w_factor - (woi[1] - woi[0]) % min_w_factor
            )
        if (woi[2]) % min_h_factor:
            woi[2] = max([woi[2] - woi[2] % min_h_factor, 0])
        if (woi[3] - woi[2] + 1) % min_h_factor:
            woi[3] = (
                woi[3] + min_h_factor - (woi[3] - woi[2]) % min_h_factor
            )

        # Set new WOI
        for i, prop in enumerate(woi_prop):
            errs.append(
                self.xeneth.XC_SetPropertyValueL(self.cam, prop, c_long(woi[i]), "")
            )

        # Report new WOI
        report_str = "     New WOI setup: "
        for i, prop in enumerate(woi_prop):
            errs.append(
                self.xeneth.XC_GetPropertyValueL(self.cam, prop, byref(prop_val))
            )
            report_str += "%s: %d | " % (prop.decode(), prop_val.value)
            woi[i] = prop_val.value
        if verbose:
            print(report_str)

        # Reconfigure buffer based on new WOI
        self.shape = (woi[3] - woi[2] + 1, woi[1] - woi[0] + 1)
        self.woi = (woi[0], self.shape[0], woi[2], self.shape[1])
        self.frame_size = self.xeneth.XC_Getframe_size(self.cam)
        self.frame_buffer = (c_ushort * int(self.frame_size / 2))(0)
        self.last_capture = np.empty(self.shape)

        if any(errs):
            print("Warning! Error(s) encountered: ", errs)

        return self.woi

    def set_low_gain(self, enable=True):
        """
        Enables or disables low gain mode.

        Parameters
        ----------
        enable : bool
            Whether to enable or disable.
        """
        gain_current = c_long(0)
        self.xeneth.XC_GetPropertyValueL(self.cam, b"LowGain", byref(gain_current))
        gain_current = bool(gain_current.value)
        if enable and not gain_current:
            print("Enabling low gain mode...")
            self.xeneth.XC_SetPropertyValueL(self.cam, b"LowGain", c_long(1), "")
        elif (not enable) and gain_current:
            print("Disabling low gain mode...")
            self.xeneth.XC_SetPropertyValueL(self.cam, b"LowGain", c_long(0), "")

    def enable_cooling(self, enable=True):
        """
        Enables/disables TEC and high fan speed.

        Parameters
        ----------
        enable : bool
            Whether to enable or disable.
        """
        fan = c_long(0)
        self.xeneth.XC_GetPropertyValueL(self.cam, b"Fan", byref(fan))
        fan_current = bool(fan.value)
        if enable and not fan_current:
            print("Enabling cooling/high fan speed...")
            self.xeneth.XC_SetPropertyValueL(self.cam, b"Fan", c_long(1), "")
        elif (not enable) and fan_current:
            print("Disabling cooling/high fan speed...")
            self.xeneth.XC_SetPropertyValueL(self.cam, b"Fan", c_long(0), "")

    ### Image Grabbing ###

    def setup(self, profile, fpt=1):
        """
        Sample pre-configured imaging profiles.

        Likely to be reconfigured by end-user for various imaging tasks.

        Parameters
        ----------
        profile
            See :attr:`profile`.
        fpt : int
            Frames per trigger for intput trigger.
        """
        if profile == "triggered":
            # 0.1 ms exposure (allows 1kHz imaging w/ 350x350 WOI w/ 400Hz Cheetah 640)
            self.set_exposure(100e-6)
            # Rising edge trigger on camera; 1 frame each trigger
            self.setup_input_trigger(mode=2, source=0, fpt=1)
            # Need 2x avoid pre-trigger
            self.setup_input_trigger(mode=2, source=0, fpt=fpt)
            # Enable low gain
            self.set_low_gain(False)
        elif profile == "free":
            # Free running, software trigger
            self.setup_input_trigger()
            # 30 ms fixed exposure
            self.set_exposure(7e-3)
            # Start free-running capture
            self.start_capture()
        else:
            print("Profile not found! Returning...")

    def snap(self, conversion=False):
        """
        Start capture, grab image, stop capture.

        Parameters
        ----------
        conversion : bool
            Makes an internal 8 bit buffer for :meth:`xeneth.SaveData` and :meth:`xeneth.Blit`
        """
        print("Starting capture...")
        err = self.xeneth.XC_StartCapture(self.cam)
        if err != I_OK:
            print("Could not start capturing, errorCode: %lu" % (err))
        elif self.xeneth.XC_IsCapturing(self.cam):
            print("Grabbing a frame...")
            if conversion:
                err = self.xeneth.XC_GetFrame(
                    self.cam,
                    FT_NATIVE,
                    XGF_BLOCKING,
                    self.frame_buffer,
                    self.frame_size,
                )
            else:  # Max performance by skipping 8 bit buffer gen
                err = self.xeneth.XC_GetFrame(
                    self.cam,
                    FT_NATIVE,
                    XGF_BLOCKING | XGF_NOCONVERSION,
                    self.frame_buffer,
                    self.frame_size,
                )
            if err != I_OK:
                print("Problem while fetching frame, errorCode %lu" % (err))
            else:
                im = np.frombuffer(self.frame_buffer, c_ushort).reshape(self.shape)
                print("Stopping capture...")
                err = self.xeneth.XC_StopCapture(self.cam)
                if err != I_OK:
                    print("Could not stop capturing, errorCode: %lu" % (err))
                else:
                    return im
        return -1

    def _get_image_hw(self, timeout_s=10, frame_type=FT_NATIVE, block=True, convert=True):
        """
        Main grabbing function; captures latest image into single frame buffer.

        Warning
        ~~~~~~~~
        ``timeout_s`` parameter is currently untested; setting it may lead to unintended behavior.

        Parameters
        ----------
        timeout_s : float
            The time in seconds to wait for frames to catch up with triggers.
        frame_type : FT_NATIVE, 0
            Sets type of frame conversion.
        block : bool
            Blocking read; waits up to ``timeout_s`` for frame.
        convert : bool
            Makes internal 8 bit buffer, set false for max performance.

        Returns
        -------
        int, numpy.ndarray
            Error code in the event of an error, otherwise the current frame.
        """

        # Update the timeout time (in ms) if different than API default
        # Note: To be renovated to only set timeout if different than current camera value...
        if timeout_s != 10 and block:
            self.set_timeout_api(int(1000 * timeout_s))

        # Set flag based on input options
        flag = XGF_BLOCKING if block else 0
        if not convert:
            flag = flag | XGF_NOCONVERSION

        ret = err = self.xeneth.XC_GetFrame(
            self.cam, frame_type, flag, self.frame_buffer, self.frame_size
        )
        if err == I_OK:
            t = time.perf_counter()
            self.last_capture = np.frombuffer(self.frame_buffer, c_ushort)
            self.last_tag = np.uint16(
                int(
                    format(self.last_capture[1], "012b")[:8]
                    + format(self.last_capture[0], "012b")[:8],
                    2,
                )
            )
            # Delete frame tag to avoid issues w/ autfocus/exposure.
            self.last_capture[:2] = 0
            self.last_capture = self.last_capture.reshape(self.shape)
            self.last_process_time = time.perf_counter() - t
            ret = self.last_capture

        return ret

    def get_frame_number(self):
        """
        Get number of captured frames since :meth:`start_capture()`.

        Returns
        ----------
        int
            Number of frames.
        """
        return self.xeneth.XC_GetFrameCount(self.cam)

    def start_capture(self):
        """Initiates the current capture run"""
        print("Starting capture...")
        err = self.xeneth.XC_StartCapture(self.cam)
        if err != I_OK:
            print("Could not start capturing, errorCode: %lu" % (err))
        while not self.isCapturing():
            print("Waiting for capture start...")
            time.sleep(0.1)

    def stop_capture(self):
        """Terminates the current capture run"""
        print("Stopping capture...")
        err = self.xeneth.XC_StopCapture(self.cam)
        if err != I_OK:
            print("Could not stop capturing, errorCode: %lu" % (err))

    def abort_capture(self):
        """Cancels any long, live frame captures"""
        print("Aborting capture...")
        err = self.xeneth.XC_SetPropertyValueE(self.cam, b"AbortExposure", b"Abort")
        if err != I_OK:
            print("Could not abort capture, errorCode: %lu" % (err))

    def flush(self):
        """See :meth:`.Camera.flush`"""
        time.sleep(0.1)  # Allow some time to grab free-running images
        err = self._get_image_hw(block=False, convert=False, return_img=False)
        while not err:
            err = self._get_image_hw(block=False, convert=False, return_img=False)

    def is_capturing(self):
        """
        Checks if currently capturing

        Returns
        -------
        bool
            ``True`` if capturing, ``False`` otherwise.
        """
        return self.xeneth.XC_IsCapturing(self.cam)

    ### Filters ###

    def autogain(self, enable=True):
        """
        Adds autogain and offset to current filter stack. Makes use of full dynamic range.

        Parameters
        ----------
        enable : bool
            Enables autogain if True.
        """
        if enable and "autogain" not in self.filters.keys():
            print("Enabling autogain...")
            tag = self.xeneth.XC_FLT_Queue(self.cam, b"AutoOffsetAndGain", "")
            self.filters["autogain"] = tag
        elif (not enable) and "autogain" in self.filters.keys():
            print("Disabling autogain...")
            self.xeneth.XC_RemImageFilter(self.cam, self.filters["autogain"])
            self.filters.pop("autogain")

    def autoexpose_xenics(self, enable=True, t_settle=0):
        """
        Adds Xenics autogain and offset filters to current filter stack.

        Makes use of the camera's full dynamic range.

        Parameters
        ----------
        enable : bool
            Enables autogain if True.
        t_settle : float
            Time to allow autoexposure to settle.
        """
        if enable and "autoexposure" not in self.filters.keys():
            print("Enabling autoexposure...")
            tag = self.xeneth.XC_FLT_Queue(self.cam, b"AutoExposure", "")
            self.xeneth.XC_FLT_SetParameter(self.cam, tag, b"Target", b"50")
            print(self.xeneth.XC_FLT_SetParameter(self.cam, tag, b"Outliers", b"0.0"))
            self.filters["autoexposure"] = tag
            t_start = time.perf_counter()
            while time.perf_counter() - t_start < t_settle:
                self._get_image_hw()

        elif (not enable) and "autoexposure" in self.filters.keys():
            print("Disabling autoexposure...")
            self.xeneth.XC_RemImageFilter(self.cam, self.filters["autoexposure"])
            self.filters.pop("autoexposure")

    def close_filters(self):
        """Deletes all current *tracked* filters in the stack."""
        errs = []
        for filter_key in self.filters:
            print("Closing %s filter..." % filter_key)
            errs.append(
                self.xeneth.XC_RemImageFilter(self.cam, self.filters[filter_key])
            )
        self.filters = {}
        if any(errs):
            print("Errors when closing filters! Codes: ", errs)



"""
Repository of common analytic phase patterns.
"""
import os
import warnings
import time
import numpy as np
try:
    import cupy as cp   # type: ignore
except ImportError:
    cp = np
from scipy import special
from math import factorial
import matplotlib.pyplot as plt

from slmsuite.misc.math import REAL_TYPES
from slmsuite.holography.toolbox import _process_grid

# Load CUDA code. This is used for cupy.RawKernels in this file and elsewhere.

def _load_cuda():
    with open(os.path.join(os.path.dirname(os.path.abspath(__file__)), "cuda.cu"), 'r') as file:
        CUDA_KERNELS = file.read()

    return CUDA_KERNELS

try:
    CUDA_KERNELS = _load_cuda()
except:
    warnings.warn("Unable to load toolbox/cuda.cu; cannot use custom GPU kernels.")
    CUDA_KERNELS = None

# Basic gratings.

def blaze(grid, vector=(0, 0)):
    r"""
    Returns a simple `blazed grating <https://en.wikipedia.org/wiki/Blazed_grating>`_,
    a linear phase ramp, toward a given vector in :math:`k`-space.

    .. math:: \phi(\vec{x}) = 2\pi \cdot \vec{k} \cdot \vec{x}

    Parameters
    ----------
    grid : (array_like, array_like) OR :class:`~slmsuite.hardware.slms.slm.SLM`
        :math:`\vec{x}`. Meshgrids of normalized :math:`\frac{x}{\lambda}` coordinates
        corresponding to SLM pixels, in ``(x_grid, y_grid)`` form.
        These are precalculated and stored in any :class:`~slmsuite.hardware.slms.slm.SLM`, so
        such a class can be passed instead of the grids directly.
    vector : (float, float)
        :math:`\vec{k}`. Blaze vector in normalized :math:`\frac{k_x}{k}` units.
        See :meth:`~slmsuite.holography.toolbox.convert_vector()`

    Returns
    -------
    numpy.ndarray
        The phase for this function.
    """
    (x_grid, y_grid) = _process_grid(grid)

    # Optimize phase construction based on context.
    if vector[0] == 0 and vector[1] == 0:
        result = np.zeros_like(x_grid)
    elif vector[1] == 0:
        result = (2 * np.pi * vector[0]) * x_grid
    elif vector[0] == 0:
        result = (2 * np.pi * vector[1]) * y_grid
    else:
        result = (2 * np.pi * vector[0]) * x_grid + (2 * np.pi * vector[1]) * y_grid

    return result


def sinusoid(grid, vector=(0, 0), shift=0, a=np.pi, b=0):
    r"""
    Returns a simple `holographic grating
    <https://en.wikipedia.org/wiki/Diffraction_grating#SR_(Surface_Relief)_gratings>`_,
    a sinusoidal grating, toward a given vector in :math:`k`-space.

    .. math:: \phi(\vec{x}) = \frac{a-b}{2} [1 + \cos(2\pi \cdot \vec{k} \cdot \vec{x} + s)] + b

    Important
    ---------
    Unlike a blazed grating :meth:`.blaze()`, power will efficiently be deflected toward
    the mirror -1st order at :math:`-\vec{k}`, by symmetry.

    Parameters
    ----------
    grid : (array_like, array_like) OR :class:`~slmsuite.hardware.slms.slm.SLM`
        :math:`\vec{x}`. Meshgrids of normalized :math:`\frac{x}{\lambda}` coordinates
        corresponding to SLM pixels, in ``(x_grid, y_grid)`` form.
        These are precalculated and stored in any :class:`~slmsuite.hardware.slms.slm.SLM`, so
        such a class can be passed instead of the grids directly.
    vector : (float, float)
        :math:`\vec{k}`. Blaze vector in normalized :math:`\frac{k_x}{k}` units.
        See :meth:`~slmsuite.holography.toolbox.convert_vector()`
    shift : float
        Radians to laterally shift the period of the grating by.
    a : float
        Value at one extreme of the sinusoid.
        Ignoring crosstalk,
        the 0th order will be minimized when ``|a-b|`` is equal to :math:`\pi`.
    b : float
        Value at the other extreme of the sinusoid.
        Defaults to zero, in which case ``a`` is the amplitude.

    Returns
    -------
    numpy.ndarray
        The phase for this function.
    """
    if vector[0] == 0 and vector[1] == 0:
        (x_grid, _) = _process_grid(grid)
        result = np.full_like(x_grid, (a-b)/2 * (1 + np.sin(shift)))
    else:
        result = (a-b)/2 * (1 + np.sin(blaze(grid, vector) + shift))

    # Add offset if provided.
    if b != 0:
        result += b

    return result


def binary(grid, vector=(0, 0), shift=0, a=np.pi, b=0, duty_cycle=.5):
    r"""
    Returns a simple binary grating toward a given vector in :math:`k`-space.

    .. math:: \phi(\vec{x}) =
        \left\{
            \begin{array}{ll}
                a, & (
                    [2\pi \cdot \vec{k} \cdot \vec{x} + s] \,\,\,\,\text{mod}\,\,\,\, 2\pi
                    ) < 2\pi*d \\
                b, & \text{ otherwise}.
            \end{array}
        \right.

    Note
    ----
    When parameters are chosen to produce integer period,
    this function uses speed optimizations **(implementation incomplete)**.
    Otherwise, this function uses ``np.mod`` on top of
    :meth:`~slmsuite.holography.toolbox.phase.blaze()` to compute gratings.

    Parameters
    ----------
    grid : (array_like, array_like) OR :class:`~slmsuite.hardware.slms.slm.SLM`
        :math:`\vec{x}`. Meshgrids of normalized :math:`\frac{x}{\lambda}` coordinates
        corresponding to SLM pixels, in ``(x_grid, y_grid)`` form.
        These are precalculated and stored in any :class:`~slmsuite.hardware.slms.slm.SLM`, so
        such a class can be passed instead of the grids directly.
    vector : (float, float)
        :math:`\vec{k}`. Blaze vector in normalized :math:`\frac{k_x}{k}` units.
        See :meth:`~slmsuite.holography.toolbox.convert_vector()`
    shift : float
        Radians to laterally shift the period of the grating by.
    a : float
        Value at one extreme of the binary grating.
    b : float
        Value at the other extreme of the binary grating.
        Defaults to zero, in which case ``a`` is the amplitude.
    duty_cycle : float
        Ratio of the period which is 'on'.

    Returns
    -------
    numpy.ndarray
        The phase for this function.
    """
    result = None

    if vector[0] == 0 and vector[1] == 0:
        (x_grid, _) = _process_grid(grid)
        phase = b
        if shift != 0:
            if np.mod(shift, 2*np.pi) < (2 * np.pi * duty_cycle):
                phase = a
        result = np.full(x_grid.shape, phase)
    elif vector[0] != 0 and vector[1] != 0:
        pass    # xor the next case.
    elif vector[0] == 0 or vector[1] == 0:
        period = 1/np.sum(vector)
        duty = period*duty_cycle

        period_int = np.rint(period)
        duty_int = np.rint(duty)

        if np.all(np.isclose(period, period_int)) and np.all(np.isclose(duty, duty_int)):
            pass    # Future: speed optimization.

    # If we have not set result, then we have to use the slow np.mod option.
    if result is None:
        result = np.where(
            np.mod(blaze(grid, vector) + shift, 2*np.pi) < (2 * np.pi * duty_cycle),
            b,
            a,
        )

    return result


# Basic lenses.

def _parse_focal_length(f):
    """Helper function to parse focal length used by `lens` and `axicon`."""
    if isinstance(f, REAL_TYPES):
        f = [f, f]
    if isinstance(f, (list, tuple, np.ndarray)):
        f = np.squeeze(f)

        if f.size != 2:
            raise ValueError("Expected two terms in focal list. Found {}.".format(f))
        if np.any(f == 0):
            raise ValueError("Cannot interpret a focal length of zero. Found {}.".format(f))

    return f


def lens(grid, f=(np.inf, np.inf)):
    r"""
    Returns a simple
    `thin parabolic lens <https://en.wikipedia.org/wiki/Thin_lens#Physical_optics>`_.

    When the focal length :math:`f` is isotropic,

    .. math:: \phi(\vec{x}) = \frac{\pi}{f}|\vec{x}|^2

    Otherwise :math:`\vec{\,f\,}` represents an elliptical lens,

    .. math:: \phi(x, y) = \pi \left[\frac{x^2}{f_x} + \frac{y^2}{f_y} \right]

    Parameters
    ----------
    grid : (array_like, array_like) OR :class:`~slmsuite.hardware.slms.slm.SLM`
        Meshgrids of normalized :math:`\frac{x}{\lambda}` coordinates
        corresponding to SLM pixels, in ``(x_grid, y_grid)`` form.
        These are precalculated and stored in any :class:`~slmsuite.hardware.slms.slm.SLM`, so
        such a class can be passed instead of the grids directly.
    f : float OR (float, float)
        Focus in normalized :math:`\frac{x}{\lambda}` units.
        Defaults to infinity (no lens).
        Scalars are interpreted as a non-cylindrical isotropic lens.
        See :meth:`~slmsuite.holography.toolbox.convert_vector` to convert depths in
        focal power units (inverse of :math:`f`) into other units and back.

    Returns
    -------
    numpy.ndarray
        The phase for this function.
    """
    (x_grid, y_grid) = _process_grid(grid)
    f = _parse_focal_length(f)

    # Optimize phase construction based on context (for speed, to avoid square, etc).
    if np.isfinite(f[0]) and np.isfinite(f[1]):
        return (np.pi / f[0]) * np.square(x_grid) + (np.pi / f[1]) * np.square(y_grid)
    elif np.isfinite(f[0]) and np.isfinite(f[1]):
        return (np.pi / f[0]) * np.square(x_grid)
    elif np.isfinite(f[1]):
        return (np.pi / f[1]) * np.square(y_grid)
    else:
        return np.zeros_like(x_grid)


def axicon(grid, f=(np.inf, np.inf), w=None):
    r"""
    Returns an `axicon <https://en.wikipedia.org/wiki/Axicon>`_ lens,
    the phase farfield for a Bessel beam. A (elliptically)-cylindrical axicon blazes
    according to :math:`\vec{k}_g = w / \vec{\,f\,} / 2` where
    :math:`w` is the radius of the axicon. With a flat input amplitude over
    :math:`[-w, w]`, this will produce a Bessel beam focussed at :math:`z = \vec{f}`.

    .. math:: \phi(\vec{x}) = 2\pi \cdot |\vec{k}_g \cdot \vec{x}|

    Parameters
    ----------
    grid : (array_like, array_like) OR :class:`~slmsuite.hardware.slms.slm.SLM`
        Meshgrids of normalized :math:`\frac{x}{\lambda}` coordinates
        corresponding to SLM pixels, in ``(x_grid, y_grid)`` form.
        These are precalculated and stored in any :class:`~slmsuite.hardware.slms.slm.SLM`, so
        such a class can be passed instead of the grids directly.
    f : float OR (float, float)
        Focal length (center of the axicon diamond) in normalized :math:`\frac{x}{\lambda}` units.
        Scalars are interpreted as a non-cylindrical isotropic axicon.
        Defaults to infinity (no axicon).
    w : float OR None
        See :meth:`~slmsuite.holography.toolbox._determine_source_radius()`.

    Returns
    -------
    numpy.ndarray
        The phase for this function.
    """
    (x_grid, y_grid) = _process_grid(grid)
    w = _determine_source_radius(grid, w)
    f = _parse_focal_length(f)

    angle = [w / f[0] / 2, w / f[1] / 2]    # Notice that this fraction is in radians.

    # Optimize phase construction based on context (for speed, to avoid sqrt, etc).
    if angle[0] == 0 and angle[1] == 0:
        return 0 * x_grid
    elif angle[0] == 0:
        return (2 * np.pi * angle[1]) * np.abs(y_grid)
    elif angle[1] == 0:
        return (2 * np.pi * angle[0]) * np.abs(x_grid)
    else:
        return (2 * np.pi) * np.sqrt(np.square(x_grid * angle[0]) + np.square(y_grid * angle[1]))


# Zernike.

ZERNIKE_INDEXING_DIMENSION = {"ansi" : 1, "noll" : 1, "fringe" : 1, "wyant" : 1, "radial" : 2}
ZERNIKE_INDEXING = ZERNIKE_INDEXING_DIMENSION.keys()
ZERNIKE_NAMES = [
    # Oth order
    "Piston",

    # 1st order
    "Vertical tilt",
    "Horizontal tilt",

    # 2nd order
    "Oblique astigmatism",
    "Defocus",
    "Vertical astigmatism",

    # 3rd order
    "Vertical trefoil",
    "Vertical coma",
    "Horizontal coma",
    "Oblique trefoil",

    # 4th order
    "Oblique quadrafoil",
    "Oblique secondary astigmatism",
    "Spherical aberration",
    "Vertical secondary astigmatism",
    "Vertical quadrafoil",

    # 5th order
    "Vertical pentafoil",
    "Vertical secondary trefoil",
    "Vertical secondary coma",
    "Horizontal secondary coma",
    "Oblique secondary trefoil",
    "Oblique pentafoil",

    # 6th order
    "Oblique hexafoil",
    "Oblique secondary quadrafoil",
    "Oblique trinary astigmatism",
    "Secondary spherical aberration",
    "Vertical trinary astigmatism",
    "Vertical secondary quadrafoil",
    "Vertical hexafoil",
]


def zernike_convert_index(indices, from_index="ansi", to_index="ansi"):
    """
    Helper function for converting between Zernike indexing conventions.

    Currently supported conventions:

    -  ``"radial"``
        The standard 2-dimensional :math:`n,l` indexing for
        `Zernike polynomials <https://en.wikipedia.org/wiki/Zernike_polynomials>`_,
        where :math:`n` is the radial index
        and :math:`l` is the azimuthal index.
        Denoted :math:`Z_n^l`.

    -  ``"ansi"``
        1-dimensional (0-indexed) `ANSI indices
        <https://en.wikipedia.org/wiki/Zernike_polynomials#OSA/ANSI_standard_indices>`_.
        **This is the default** :mod:`slmsuite` **index.**
        Denoted :math:`Z_i`.

    -  ``"noll"``
        1-dimensional (1-indexed) `Noll indices
        <https://en.wikipedia.org/wiki/Zernike_polynomials#Noll's_sequential_indices>`_.

    -  ``"fringe"``
        1-dimensional (1-indexed) `Fringe indices
        <https://en.wikipedia.org/wiki/Zernike_polynomials#Fringe/University_of_Arizona_indices>`_.

    -  ``"wyant"``
        1-dimensional (0-indexed) `Wyant indices
        <https://en.wikipedia.org/wiki/Zernike_polynomials#Wyant_indices>`_.
        Equivalent to ``"fringe"``, except with starting with zero instead of one.

    Parameters
    ----------
    indices : array_like
        List of indices of shape ``(N, D)`` where ``D`` is the dimension of the indexing
        (1, apart from ``"radial"`` indexing which has a dimension of 2).
    from_index, to_index : str
        Zernike index convention. Must be supported.

    Returns
    -------
    indices_converted : numpy.ndarray
        List of indices of shape ``(N, D)`` where ``D`` is the dimension of the indexing
        (1, apart from ``"radial"`` indexing which has a dimension of 2).

    Raises
    ------
    ValueError
        If an invalid index number or index type is given,
        or an invalid indices shape is given.
    """
    # Parse arguments.
    if from_index not in ZERNIKE_INDEXING:
        raise ValueError(
            f"From index '{from_index}' not recognized as a valid unit. "
            f"Options: {ZERNIKE_INDEXING}."
        )
    if to_index not in ZERNIKE_INDEXING:
        raise ValueError(
            f"To index '{to_index}' not recognized as a valid unit. "
            f"Options: {ZERNIKE_INDEXING}."
        )

    dimension = ZERNIKE_INDEXING_DIMENSION[from_index]

    indices = np.array(indices, dtype=int, copy=(False if np.__version__[0] == '1' else None))
    if indices.size == dimension:
        indices = indices.reshape((1, dimension))
    if dimension > 1 and indices.shape[1] != dimension:
        raise ValueError(f"Expected dimension (N, {dimension}); found {indices.shape}")

    if from_index == to_index:
        return indices

    # Convert all cases to radial indices n, l.
    if from_index == "radial":
        n = indices[:,0]
        l = indices[:,1]
    elif from_index == "noll" or to_index == "fringe" or from_index == "wyant":
        # Inverse functions have not been implemented.
        raise NotImplementedError(f"from_index '{from_index}' is not supported currently.")
    elif from_index == "ansi":
        n = np.floor(.5 * np.sqrt(8*indices + 1) - .5).astype(int)
        l = 2*indices - n*(n+2)

    # Error check n,l
    if np.any((n + l) % 2):
        raise ValueError(f"Invalid Zernike index n,l. n+l must be even. n={n}, l={l}.")
    if np.any(np.abs(l) > n):
        raise ValueError(f"Invalid Zernike index n,l. |l| cannot be larger than n. n={n}, l={l}.")
    if np.any(n < 0):
        raise ValueError(f"Invalid Zernike index n,l. n must be non-negative. n={n}, l={l}.")

    # Convert to the desired indices.
    if to_index == "radial":
        result = np.vstack((n, l)).T
    elif to_index == "noll":
        result = (n * (n + 1)) // 2 + np.abs(l)
        result += np.logical_and(l >= 0, np.mod(n, 4) <= 1)
        result += np.logical_and(l <= 0, np.mod(n, 4) > 1)
    elif to_index == "wyant" or to_index == "fringe":
        result = (
            np.square(1 + (n + np.abs(l)) / 2).astype(int)
            - 2 * np.abs(l) + (l < 0)
            - (to_index == "wyant")
        )
    elif to_index == "ansi":
        result = (n * (n + 2) + l) // 2

    return result


def zernike_aperture(grid, aperture=None):
    """
    Helper function to find the appropriate scaling for between the normalized units in
    the grid and the Zernike aperture (the unit disk).

    Tip
    ~~~
    Passing an :class:`~slmsuite.hardware.slms.slm.SLM` for ``grid`` makes this easy.
    The function :meth:`~slmsuite.hardware.slms.slm.SLM.get_source_zernike_scaling()`
    determines the optimal scaling of the aperture.

    Important
    ~~~~~~~~~
    Zernike polynomials are canonically defined on a circular aperture. However, we may
    want to use these polynomials on other apertures (e.g. a rectangular SLM).
    Cropping this aperture breaks the orthogonality and normalization of the set, but
    this is fine for many applications. While it is possible to orthonormalize the
    cropped set, we do not do so in :mod:`slmsuite`, as this is not critical for target
    applications such as aberration correction.

    Caution
    ~~~~~~~
    Anisotropic Zernike scaling can lead to unexpected behavior.
    For instance, the :math:`Z_4 = Z_2^0 = 1 - 2x^2 - 2y^2` Zernike term is commonly
    used for focusing, but with anisotropic scaling, this becomes an elliptical lens
    on the SLM which may not behave as expected.

    Parameters
    ----------
    aperture : {"circular", "elliptical", "cropped"} OR (float, float) OR float OR None
        How to scale the polynomials relative to the grid shape. This is relative
        to the :math:`R = 1` edge of a standard Zernike pupil.

        - ``None``
          If a :class:`~slmsuite.hardware.slms.slm.SLM` is passed for ``grid``, then
          uses :meth:`~slmsuite.hardware.slms.slm.SLM.get_source_zernike_scaling()` to
          determine the scaling most appropriate for the SLM.
          Otherwise, defaults to ``"cropped"``.
          See also :meth:`~slmsuite.hardware.slms.slm.SLM.fit_source_amplitude()`
          and especially the `extent_threshold` keyword which determines the scaling used by
          :meth:`~slmsuite.hardware.slms.slm.SLM.get_source_zernike_scaling()`

        - ``"circular"``
          The circle is scaled isotropically until the pupil edge touches one set
          of opposite grid edges. This is the default aperture.

        - ``"elliptical"``
          The circle is scaled anisotropically until each pupil edge touches a grid
          edge. Generally produces an ellipse.

        - ``"cropped"``
          The circle is scaled isotropically until the rectangle of the grid is
          circumscribed by the circle.

        - ``float OR (float, float)``
          Custom scaling. These values are multiplied to the ``x_grid`` and ``y_grid``
          directly, respectively. The edge of the Zernike pupil corresponds to where
          ``(s_x * x_grid)**2 + (s_y * y_grid)**2 = 1``. If a scalar is given, assumes
          isotropic scaling.

    Returns
    ~~~~~~~
    (float, float)
    """
    # Parse grid.
    (x_grid, y_grid) = _process_grid(grid)

    # Parse aperture.
    if aperture is None:
        # Check if cameraslm.
        if hasattr(grid, "slm") and hasattr(grid, "cam"):
            grid = grid.slm

        # Check if slm.
        if hasattr(grid, "get_source_zernike_scaling"):
            aperture = grid.get_source_zernike_scaling()
        else:
            aperture = "cropped"

    if isinstance(aperture, str):
        if aperture == "elliptical":
            x_scale = 1 / np.nanmax(x_grid)
            y_scale = 1 / np.nanmax(y_grid)
        elif aperture == "circular":
            x_scale = y_scale = 1 / np.amin([np.nanmax(x_grid), np.nanmax(y_grid)])
        elif aperture == "cropped":
            x_scale = y_scale = 1 / np.sqrt(np.nanmax(np.square(x_grid) + np.square(y_grid)))
        else:
            raise ValueError(f"Aperture '{aperture}' is not implemented.")
    elif np.isscalar(aperture):
        x_scale = y_scale = aperture
    elif isinstance(aperture, (list, tuple, np.ndarray)) and len(aperture) == 2:
        x_scale = aperture[0]
        y_scale = aperture[1]
    else:
        raise ValueError("Aperture type {} not recognized.".format(type(aperture)))

    return (x_scale, y_scale)


def zernike(grid, index, weight=1, **kwargs):
    r"""
    Returns a single real `Zernike polynomial <https://en.wikipedia.org/wiki/Zernike_polynomials>`_.
    These polynomials are commonly used as an orthonormal basis for optical aberration
    and are used in a number of places inside :mod:`slmsuite` for aberration compensation.

    Parameters
    ----------
    grid : (array_like, array_like) OR :class:`~slmsuite.hardware.slms.slm.SLM`
        Meshgrids of normalized :math:`\frac{x}{\lambda}` coordinates
        corresponding to SLM pixels, in ``(x_grid, y_grid)`` form.
        These are precalculated and stored in any :class:`~slmsuite.hardware.slms.slm.SLM`, so
        such a class can be passed instead of the grids directly.
    index : int
        ANSI Zernike index defining the polynomial.
    weight : float
        Amplitude of the polynomial.
    **kwargs
        Passed to :meth:`.zernike_sum()`.

    Returns
    -------
    numpy.ndarray
        The phase for this function.
    """
    return zernike_sum(grid, (int(index),), (float(weight),), **kwargs)


def zernike_get_string(index, derivative=(0,0)):
    r"""
    Returns a :math:`\LaTeX` string corresponding to the cartesian representation of the
    Zernike polynomial of the given index. The monomials are sorted in reverse Cantor order.

    Parameters
    ----------
    index : int
        ANSI Zernike index.
    derivative : (int, int)
        If non-negative, returns the Zernike derivative of the given order. For instance,
        ``(1, 0)`` corresponds to the first derivative in the :math:`x` direction.
    """
    cxy, cw = _zernike_get_cantor([index], [[1]], derivative)
    result = ""

    # Sum the monomial terms together.
    for i, w in zip(reversed(range(len(cw))), reversed(cw[:, 0])):
        result += "{0:+}".format(int(w))

        for j, n in enumerate(["x", "y"]):
            if cxy[i, j] >= 1:
                result += n

                if cxy[i, j] > 1:
                    result += f"^{cxy[i, j]}"

    if len(result) == 0:
        result = "0"

    return result.strip("+")    # Remove potential leading +


def _zernike_get_cantor(indices, weights, derivative=(0,0)):
    indices = np.array(indices)
    weights = np.array(weights)

    # Separate the negative indices (special cases) before processing.
    negative_mask = indices < 0
    positive_mask = indices >= 0

    negative_indices = indices[negative_mask]
    indices = indices[positive_mask]

    negative_weights = weights[negative_mask, :]
    weights = weights[positive_mask, :]

    # Grab the zernike-cantor transformation from the cache.
    _zernike_build_indices(indices)
    zernike_cantor = _zernike_cache_vectorized[indices, :]   # (D, M)
    M = zernike_cantor.shape[1]
    cantor_indices = np.arange(M)

    # Remove vectors with all zeros.
    nonzero = np.any(zernike_cantor, axis=0)    # Which D are nonzero for given m in M
    cantor_indices = cantor_indices[nonzero]    # M -> M'
    zernike_cantor = zernike_cantor[:, nonzero] # (D, M')

    cantor_pairing = _inverse_cantor_pairing(cantor_indices)    # (M', 2)

    # Differentiate the terms if needed.
    if np.any(derivative):

        for j in [0, 1]:
            if derivative[j] > 0:
                power = cantor_pairing[:, [j]].T.astype(int)  # (D, 1)

                # Apply the power rule.
                if derivative[j] == 1:
                    zernike_cantor *= power
                elif derivative[j] > 1:
                    nonzero = power >= derivative[j]

                    zernike_cantor[np.logical_not(nonzero)] = 0
                    zernike_cantor[nonzero] *= (
                        special.factorial(power[nonzero]) / special.factorial(power[nonzero] - derivative[j])
                    ).astype(int)

                # Reduce the power of the term
                cantor_pairing[:, j] -= derivative[j]
                cantor_pairing[cantor_pairing[:, j] < 0, j] = 0
                zernike_cantor *= power >= derivative[j]

        # Remove terms with all zeros
        nonzero = np.any(zernike_cantor, axis=0)        # Which D are nonzero for given m in M'
        cantor_pairing = cantor_pairing[nonzero, :]     # M' -> M''
        zernike_cantor = zernike_cantor[:, nonzero]     # (D, M'')

    # Reshape the weights into this new basis.
    cantor_weights = np.matmul(zernike_cantor.T, weights)  # (M' or M'', D) x (D, N) = (M' or M'', N)

    # Add in the negative indices.
    (M, N) = cantor_weights.shape
    MM = M + np.sum(negative_mask)

    final_pairing = np.zeros((MM, 2), dtype=int)
    final_pairing[:M, :] = cantor_pairing
    final_pairing[M:, 0] = negative_indices

    final_weights = np.zeros((MM, N))
    final_weights[:M, :] = cantor_weights
    final_weights[M:, :] = negative_weights

    return final_pairing, final_weights


def _zernike_indices_parse(indices=None, D=None, smaller_okay=False):
    """
    Parse Zernike indices applied to data expecting size D.
    """
    # Deal with the scalar case: a request for DD indices.
    if np.isscalar(indices):
        DD = int(indices)
        if D is None:
            if not smaller_okay:
                D = DD
        elif not ((smaller_okay and D <= DD) or D == DD):
            raise ValueError(f"Expected data (dimension {D}) to have common size with indices (requested {DD}).")

        D = DD

        # Fill in indices based on D now.
        indices = None

    # If None, assume list based on D.
    if indices is None:
        if D is None:
            raise ValueError("Either dimension or indices must be defined.")
        elif D == 2:
            indices = np.array([2,1])
        elif D == 3:
            indices = np.array([2,1,4])
        elif D == 4:
            indices = np.array([2,1,4,3])
        else:
            indices = np.hstack((np.array([2,1,4,3]), np.arange(5, D+1)))

    # Final checks.
    indices = np.ravel(indices)
    if indices.ndim == 0:
        indices = np.array([indices])
    if D is not None and not ((smaller_okay and D <= len(indices)) or D == len(indices)):
        raise ValueError(f"Expected data (dimension {D}) to have common size with indices (length {len(indices)}).")

    return indices


def zernike_sum(grid, indices, weights, aperture=None, use_mask=True, derivative=(0,0), out=None):
    r"""
    Returns a summation of
    `Zernike polynomials <https://en.wikipedia.org/wiki/Zernike_polynomials>`_
    in a computationally-efficient manner.
    These polynomials are commonly used as an orthonormal basis for optical aberration
    and are used in a number of places inside :mod:`slmsuite` for aberration compensation.
    This function returns a sum of polynomials:

    .. math:: \phi(\vec{x}) = \sum_i w_i Z_{I_i}(\vec{x}).

    where :math:`I_i` are the ANSI ``indices`` of the polynomials and
    :math:`w_i` are the floating point ``weights`` of each polynomial.
    To improve performance, especially for higher order polynomials,
    we store a cache of Zernike coefficients to avoid regeneration.
    See the below example to generate
    :math:`Z_1 - Z_2 + Z_4 = Z_1^{-1} - Z_1^1 + Z_2^0`,
    where the standard radial Zernike indexing :math:`Z_n^l`
    is instead represented as :math:`Z_j` by the 1-dimensional ANSI index :math:`j`.

    .. highlight:: python
    .. code-block:: python

        zernike_sum_phase = toolbox.phase.zernike_sum(
            grid=slm,
            indices=(1,  2,  4),    # Define Z_1, Z_2, Z_4
            weights=(1, -1,  1),    # Request Z_1 - Z_2 + Z_4
            aperture="circular"
        )

    Parameters
    ----------
    grid : (array_like, array_like) OR :class:`~slmsuite.hardware.slms.slm.SLM`
        Meshgrids of normalized :math:`\frac{x}{\lambda}` coordinates
        corresponding to SLM pixels, in ``(x_grid, y_grid)`` form.
        These are precalculated and stored in any :class:`~slmsuite.hardware.slms.slm.SLM`, so
        such a class can be passed instead of the grids directly.
    indices : array_like of int OR None
        Which Zernike polynomials to sum, defined by ANSI indices. Of shape ``(D,)``.

        Use :meth:`~slmsuite.holography.toolbox.phase.zernike_convert_index()`
        to convert to ANSI from various other common indexing conventions.

        If ``None`` is passed, the assumed Zernike basis depends on the
        dimensionality of the provided spots:

        -   If ``D == 2``, then the basis is assumed to be ``[2,1]``
            corresponding to the :math:`x = Z_2 = Z_1^1`
            and :math:`y = Z_1 = Z_1^{-1}` tilt terms.

        -   If ``D == 3``, then the basis is assumed to be ``[2,1,4]``
            corresponding to the previous, with the addition of the
            :math:`Z_4 = Z_2^0` focus term.

        -   If ``D > 3``, then the basis is assumed to be ``[2,1,4,3,5,6...,D]``.
            The piston term (Zernike index 0) is ignored as this constant phase is
            not relevant.

    weights : array_like of float
        The weight for each given index. Of shape ``(D,)``.
        If a stack of zernike sums is desired, then use shape ``(D, N)``.
    aperture : {"circular", "elliptical", "cropped"} OR (float, float) OR float OR None
        Determines how the Zernike polynomials are laterally scaled.
        Parsed with :meth:`~slmsuite.holography.toolbox.phase.zernike_aperture()`.

        Important
        ~~~~~~~~~
        Read the documentation and tips in
        :meth:`~slmsuite.holography.toolbox.phase.zernike_aperture()`
        to avoid subtle issues with lateral scaling.

    use_mask : bool OR "return"
        If ``True``, sets the area where standard Zernike polynomials are undefined to zero.
        If ``False``, the polynomial is not cropped. This should be used carefully, as
        the wavefront correction outside the unit circle quickly explodes with
        :math:`r^O` for terms of high order :math:`O`.
        If ``"return"``, returns the 2D mask ``x_grid**2 + y_grid**2 <= 1``.
    derivative : (int, int)
        If non-negative, returns the Zernike derivative of the given order. For instance,
        ``(1, 0)`` corresponds to the first derivative in the :math:`x` direction.
        This is fast and accurate because the derivative is computed via power rule before
        generating Zernike images.
    out : array_like OR None
        Memory to be used for the phase output. Allocated separately if ``None``.

    Returns
    -------
    numpy.ndarray
        The phase for this function. Optionally returns the 2D Zernike mask.
    """
    # Parse passed simple values.
    (x_grid, y_grid) = _process_grid(grid)
    (x_scale, y_scale) = zernike_aperture(grid, aperture)
    if len(derivative) != 2:
        raise ValueError("Expected derivative to be a (int, int)")

    # Parse weights.
    weights = np.squeeze(weights)
    if weights.ndim <= 1:
        if weights.ndim == 0:
            weights = np.array([weights])

        if indices is None:
            D = None
        else:
            indices = np.squeeze(indices)
            if indices.ndim == 0:
                indices = np.array([indices])

            D = len(indices)

        if D is None or len(weights) == D:
            weights = np.reshape(weights, (-1, 1))
        else:
            raise ValueError("Expected weights to have a common dimension with indices.")
    elif weights.ndim == 2:
        pass
    else:
        raise ValueError("Expected weights to be 1D or 2D.")

    (D, N) = weights.shape

    # Parse indices.
    indices = _zernike_indices_parse(indices, D)

    # Parse out.
    out = _parse_out(x_grid, out, stack=N)

    # At the end, we're going to set the values outside the aperture to zero.
    # Make a mask for this if it's necessary.
    if use_mask is False:
        mask = None
    else:
        mask = np.square(x_grid * x_scale) + np.square(y_grid * y_scale) <= 1
        if use_mask == "return":
            return mask
        use_mask = use_mask and np.any(mask == 0)

    # Make the new grids.
    if use_mask:
        x_grid_scaled = x_grid[mask] * x_scale
        y_grid_scaled = y_grid[mask] * y_scale
    else:
        # Special case to avoid copying grids in the case of no scaling.
        if x_scale == 1:    x_grid_scaled = x_grid
        else:               x_grid_scaled = x_grid * x_scale
        if y_scale == 1:    y_grid_scaled = y_grid
        else:               y_grid_scaled = y_grid * y_scale

    # Gather the Zernike information.
    cantor_terms, cantor_weights = _zernike_get_cantor(indices, weights, derivative)

    # The masked case only computes on a fraction of the full space.
    if use_mask:
        out[:, mask] = polynomial(
            grid=(x_grid_scaled, y_grid_scaled),
            weights=cantor_weights,
            terms=cantor_terms,
            out=out[:, mask]
        )
    else:
        out = polynomial(
            grid=(x_grid_scaled, y_grid_scaled),
            weights=cantor_weights,
            terms=cantor_terms,
            out=out
        )

    if N == 1:
        return out.reshape(x_grid.shape)
    else:
        return out


def zernike_pyramid_plot(
        grid,
        order,
        scale=1,
        titles=["ansi", "radial", "latex", "name"],
        cmap="twilight_shifted",
        **kwargs
    ):
    r"""
    Plots :meth:`.zernike()` on a pyramid of subplots corresponding to the radial and
    azimuthal order. The user can resize the figure with ``plt.figure()`` beforehand
    and force ``plt.show()`` afterward.

    Parameters
    ----------
    grid : (array_like, array_like) OR :class:`~slmsuite.hardware.slms.slm.SLM`
        Meshgrids of normalized :math:`\frac{x}{\lambda}` coordinates
        corresponding to SLM pixels, in ``(x_grid, y_grid)`` form.
        These are precalculated and stored in any :class:`~slmsuite.hardware.slms.slm.SLM`, so
        such a class can be passed instead of the grids directly.
    order : int
        Maximum radial order to plot.
    scale : float
        Scales the subplots to ``[-scale, scale]``.
    titles : list of str
        Which titles to plot. Options:

        -   ``"ansi"`` the ANSI singleton index,
        -   ``"radial"`` the radial index pair,
        -   ``"latex"`` the cartesian representation of the polynomial,
        -   ``"name"`` the name of the aberration produced by the polynomial.
    cmap : str
        Colormap to use in plotting.
    **kwargs
        Passed to :meth:`.zernike()`.
    """
    order = int(order + 1)
    indices_ansi = np.arange((order * (order + 1)) // 2)
    indices_radial = zernike_convert_index(indices_ansi, from_index="ansi", to_index="radial")
    derivative = kwargs["derivative"] if "derivative" in kwargs else (0,0)

    # Get the pitch of the subplots for later.
    a1 = plt.subplot(order, order, 1)
    a2 = plt.subplot(order, order, 2)

    pitch = a2.get_position().xmin - a1.get_position().xmin

    a1.remove()
    a2.remove()

    # Grab all the phases as a stack.
    grid_ = _process_grid(grid)
    phases = np.zeros((len(indices_ansi), *grid_[0].shape))

    phases = zernike_sum(grid, indices_ansi[np.newaxis, :], np.diag(np.ones_like(indices_ansi)), out=phases, **kwargs)

    for i in indices_ansi:
        n, l = indices_radial[i, :]
        m = (n + l) // 2

        a = plt.subplot(order, order, 1 + m + n*order)

        # Plot the phase.
        plt.imshow(phases[i], cmap=cmap)

        # Construct the title.
        title = ""

        if "ansi" in titles:
            title += f"#{i}\n"
        if "radial" in titles:
            title += f"({n}, {l})\n"
        if "latex" in titles:
            latex = zernike_get_string(i, derivative)
            title += "$" + latex + "$\n"
        if derivative == (0,0) and "name" in titles and i < len(ZERNIKE_NAMES):
            title += ZERNIKE_NAMES[i]

        plt.title(title.strip("\n"))

        # Set scales and move the axis.
        plt.clim([-scale, scale])
        plt.xticks([])
        plt.yticks([])

        dx = .5 * (3*order - n)
        box = a.get_position()
        box = box.translated(dx * pitch, 0)
        a.set_position(box)


def _zernike_cache_plot():
    plt.figure(figsize=(10,10))
    plt.imshow(np.log2(_zernike_cache_vectorized))
    plt.ylabel("Zernike Index (ANSI)");
    plt.xlabel("Monomial Index (Cantor)");


# Old style dictionary.     {(n,m) : {(nx, ny) : w, ... }, ... }
_zernike_cache = {}

# New style matrix.         N x M, N: ANSI Zernike, M: cantor polynomial.
_zernike_cache_vectorized = np.array([[]], dtype=int)


def _zernike_build_order(o):
    """Pre-caches Zernike polynomial coefficients up to order :math:`n`."""
    N = (n+1) * (n+2) // 2
    for i in range(N):
        _zernike_coefficients(i)


def _zernike_build_indices(indices):
    """Pre-caches Zernike polynomial coefficients up to order :math:`n`."""
    for i in indices:
        _zernike_coefficients(i)


def _zernike_coefficients(index):
    """
    Returns the coefficients for the :math:`x^ay^b` terms of the real Zernike polynomial
    of ANSI index ``i``. This is returned as a dictionary of form ``{(a,b) : coefficient}``.
    Uses `this algorithm <https://doi.org/10.1117/12.294412>`_.
    """
    index = int(index)

    # Generate coefficients only if we have not already generated.
    if not index in _zernike_cache:
        zernike_this = {}

        (n, l) = zernike_convert_index(index, to_index="radial")[0]
        l = -l

        # Define helper variables.
        if l % 2:   # If even
            q = int((abs(l) - 1) / 2)
        else:
            if l > 0:
                q = int(abs(l)/2 - 1)
            else:
                q = int(abs(l)/2)

        if l <= 0:
            p = 0
        else:
            p = 1

        l = abs(l)
        m = int((n-l)/2)

        # Helper function
        def comb(n, k):
            return factorial(n) / (factorial(k) * factorial(n-k))

        # Finding the coefficients is a summed combinatorial search.
        # This is why we cache: so we don't have to do this many times,
        # especially for higher order polynomials and the corresponding cubic scaling.
        for i in range(q+1):
            for j in range(m+1):
                for k in range(m-j+1):
                    factor = -1 if (i + j) % 2 else 1
                    factor *= comb(l, 2 * i + p)
                    factor *= comb(m - j, k)
                    factor *= (
                        float(factorial(n - j))
                        / (factorial(j) * factorial(m - j) * factorial(n - m - j))
                    )

                    power_key = (int(n - 2*(i + j + k) - p), int(2 * (i + k) + p))

                    # Add this coefficient to the element in the dictionary
                    # corresponding to the right power.
                    if power_key in zernike_this:
                        zernike_this[power_key] += int(factor)
                    else:
                        zernike_this[power_key] = int(factor)

        # Update the cache. Remove all factors that have cancelled out (== 0).
        _zernike_cache[index] = {
            power_key: factor
            for power_key, factor in zernike_this.items()
            if factor != 0
        }

        # If we need to, enlarge the vector cache.
        N = (n+1) * (n+2) // 2      # The Zernike order determines the size of the cache.
        global _zernike_cache_vectorized

        if _zernike_cache_vectorized.shape[1] < N:
            _zernike_cache_vectorized = np.pad(
                _zernike_cache_vectorized,
                (
                    (0, N - _zernike_cache_vectorized.shape[0]),
                    (0, N - _zernike_cache_vectorized.shape[1])
                ),
                constant_values=0
            )

        # Update the vectorized dict.
        for power_key, factor in _zernike_cache[index].items():
            cantor_index = _cantor_pairing(power_key)
            _zernike_cache_vectorized[index, cantor_index] = factor

    return _zernike_cache[index]



def _zernike_populate_basis_map(indices):
    """
    This generates helper maps ``c_md``, ``i_md``, ``pxy_m`` for use in GPU kernels
    (see ``populate_basis`` in cuda.cu).
    """
    indices = np.squeeze(indices)
    D = len(indices)

    # Omit negative indices (special cases)
    zernike_indices = indices[indices >= 0]
    other_indices = indices[indices < 0]

    # Make sure all coefficients are generated.
    for i in zernike_indices:
        _zernike_coefficients(i)

    # Determine the cantor indices.
    nonzero_cantor_indices = np.any(_zernike_cache_vectorized[zernike_indices, :], axis=0)
    cantor_indices = np.arange(len(nonzero_cantor_indices), dtype=int)[nonzero_cantor_indices]

    M = len(cantor_indices)

    pxy_m = _inverse_cantor_pairing(cantor_indices).astype(np.int32)

    # Find an optimal sort pattern for constructing the polynomials.
    # msort = _term_pathing(pxy_m)
    msort = np.arange(M)
    pxy_m = pxy_m[msort, :]

    # Reinsert the other cases.
    if len(other_indices) > 0:
        pxy_m = np.pad(pxy_m, ((0, len(other_indices)), (0,0)))
        pxy_m[len(zernike_indices):, 0] = other_indices     # Other indices go into nx.

    # Populate the results.
    c_md = _zernike_cache_vectorized[zernike_indices, :][:, cantor_indices[msort]].T.astype(np.float32)
    i_md = np.full((M, D), -1, dtype=np.int32)

    darange = np.arange(len(zernike_indices))

    for m in msort:
        nonzero = darange[c_md[m, :] != 0]
        i_md[m, :len(nonzero)] = nonzero

    return c_md, i_md, pxy_m.T


try:
    _zernike_test_kernel = cp.RawKernel(CUDA_KERNELS, 'zernike_test')
except:
    _zernike_test_kernel = None


def _zernike_test(grid, indices):
    _zernike_test_kernel = cp.RawKernel(_load_cuda(), 'zernike_test')

    c_md, i_md, pxy_m = _zernike_populate_basis_map(indices)

    # Parse grid.
    (x_grid, y_grid) = _process_grid(grid)
    scale = 1
    if hasattr(grid, "get_source_zernike_scaling"):
        scale = grid.get_source_zernike_scaling()
    x_grid = cp.array(x_grid * scale, copy=(False if np.__version__[0] == '1' else None), dtype=np.float32)
    y_grid = cp.array(y_grid * scale, copy=(False if np.__version__[0] == '1' else None), dtype=np.float32)

    (H, W) = x_grid.shape
    WH = int(W*H)
    (M, D) = c_md.shape

    out = cp.empty((D,H,W), dtype=np.float32)

    threads_per_block = int(_zernike_test_kernel.max_threads_per_block)
    blocks = WH // threads_per_block

    # Call the RawKernel.
    _zernike_test_kernel(
        (blocks,),
        (threads_per_block,),
        (
            WH, D, M,
            cp.array(c_md.ravel(), copy=(False if np.__version__[0] == '1' else None)),
            cp.array(i_md.ravel(), copy=(False if np.__version__[0] == '1' else None)),
            cp.array(pxy_m.ravel(), copy=(False if np.__version__[0] == '1' else None)),
            x_grid.ravel(),
            y_grid.ravel(),
            out.ravel()
        )
    )

    return out


# Polynomials.

def _cantor_pairing(xy):
    """
    Converts a 2D index to a unique 1D index according to the
    `Cantor pairing function <https://en.wikipedia.org/wiki/Pairing_function>`.
    """
    xy = np.array(xy, dtype=int, copy=(False if np.__version__[0] == '1' else None)).reshape((-1, 2))
    return np.rint(.5 * (xy[:,0] + xy[:,1]) * (xy[:,0] + xy[:,1] + 1) + xy[:,1]).astype(int)


def _inverse_cantor_pairing(z):
    """
    Converts a 1D index to a unique 2D index according to the
    `Cantor pairing function <https://en.wikipedia.org/wiki/Pairing_function>`.

    Returns shape ``(D, 2)``
    """
    z = np.array(z, dtype=int, copy=(False if np.__version__[0] == '1' else None))
    if z.ndim != 1:
        raise ValueError("Expected a list of shape (D,)")

    w = np.floor((np.sqrt(8*z + 1) - 1) // 2).astype(int)
    t = (w*w + w) // 2

    y = z-t
    x = w-y

    # Handle negative index case which is used for special indices.
    y[z < 0] = 0
    x[z < 0] = z[z < 0]

    return np.vstack((x, y)).T


def _term_pathing(xy):
    """
    Returns the index for term sorting to minimize number of monomial multiplications when summing
    polynomials (with only one storage variable).

    It may be the case that division could yield a shorter path, but division is
    generally more expensive than multiplication so we omit this scenario.

    It may also be the case that optimizing for large-step multiplications can yield a
    speedup. (e.g. `x^5 = y * y * x` with `y = x * x` costs three multiplications instead
    of five) However, it is unlikely that users will need the very-high-order
    polynomials would would experience an appreciable speedup.

    Parameters
    ----------
    xy : array_like
        Array of shape ``(M, 2)``.

    Returns
    -------
    I : numpy.ndarray
        Array of shape ``(M,)``. Best coefficient order.
    """
    # Prepare helper variables.
    xy = np.array(xy, dtype=int, copy=(False if np.__version__[0] == '1' else None))

    order = np.sum(xy, axis=1)
    delta = np.squeeze(np.diff(xy, axis=1))

    cantor = _cantor_pairing(xy)
    cantor_index = np.argsort(-cantor)

    # Prepare the output data structure.
    I = np.zeros_like(order, dtype=int)

    # Helper function to recurse through pathing options.
    def recurse(i0, j0):
        # Fill in the current values.
        I[j0] = cantor_index[i0]
        cantor[cantor_index[i0]] = -1

        if j0 == 0:
            return 0

        # Figure out the distance between the current index and all other indices.
        dd = delta - delta[cantor_index[i0]]
        do = order[cantor_index[i0]] - order

        # Find the best candidate for the next index in the thread.
        nearest = -cantor + np.where((np.abs(dd) > do) + (do <= 0) + (cantor < 0), np.inf, 0)
        i = np.argmin(nearest[cantor_index])

        # Either exit or continue this thread.
        if cantor[cantor_index[i]] != -1:
            return recurse(i, j0-1)
        else:
            return j0-1

    # Traverse backwards through the array,
    j = len(I)-1
    for i in range(len(order)):
        if cantor[cantor_index[i]] >= 0 and j >= 0:
            j = recurse(i, j)

    return I


def _parse_out(x_grid, out, stack=1):
    """
    Helper function to error check the shape and type of ``out``.
    """
    shape = tuple(np.concatenate(([stack], x_grid.shape)))

    if out is None:
        # Initialize out to zero.
        if cp == np:
            out = np.zeros(shape, x_grid.dtype)
        else:
            out = cp.get_array_module(x_grid).zeros(shape, x_grid.dtype)

        return out
    else:
        # Error check user-provided out.
        if out.size != np.prod(shape):
            raise ValueError("out must have same size as the stacked grid.")
        if out.dtype != x_grid.dtype:
            raise ValueError("out must have same type as grid.")
        if cp != np and cp.get_array_module(x_grid) != cp.get_array_module(out):
            raise ValueError("out and grid must both be cupy arrays if one is.")

        return out.reshape(shape)


def polynomial(grid, weights, terms=None, pathing=None, out=None):
    r"""
    Returns a summation of monomials. Specifically,

    .. math:: \phi(x, y) = \sum_{n,m \in T} w_{nm}x^ny^m

    where :math:`w_{nm}` are floating-point weights.

    Parameters
    ----------
    grid : (array_like, array_like) OR :class:`~slmsuite.hardware.slms.slm.SLM`
        Meshgrids of normalized :math:`\frac{x}{\lambda}` coordinates
        corresponding to SLM pixels, in ``(x_grid, y_grid)`` form.
        These are precalculated and stored in any :class:`~slmsuite.hardware.slms.slm.SLM`, so
        such a class can be passed instead of the grids directly.
    weights : array_like of float
        Array of shape ``(D,)`` corresponding to the coefficient of each term.
        Can also be shape ``(D, N)`` if a stack of ``N`` polynomials is desired.
    terms : array_like of int OR None
        Array of shape ``(D, 2)`` corresponding to the :math:`x` and :math:`y` exponents
        for the ``D`` terms.
        Otherwise, array of shape ``(D,)`` corresponding to the Cantor indices of
        monomials.
        If ``None``, assumes the terms are Cantor indices of the range of ``weights``.
    pathing : array_like of int OR None
        Array of shape ``(D,)`` corresponding to an order that the terms should be
        calculated. If ``None``, chooses the path that reduces the number of
        multiplications when evaluating monomials.
    out : numpy.ndarray OR cupy.ndarray
        A location where the result is stored. Use this to avoid allocating new memory.

    Returns
    -------
    out : numpy.ndarray OR cupy.ndarray
        Result of the sum.
    """
    # Parse terms
    terms = np.array(terms)
    if terms is None:
        terms = _inverse_cantor_pairing(np.arange(D))

    if terms.ndim == 1:
        terms = _inverse_cantor_pairing(terms)

    if terms.shape[1] != 2:
        raise ValueError("Terms must be of shape (D, 2) or (D,). Found {}.".format(terms.shape))

    D = terms.shape[0]

    # Parse weights
    weights = np.array(weights)
    if weights.ndim == 1:
        if len(weights) == D:
            weights = np.reshape(weights, (-1, 1))
        else:
            raise ValueError("Expected weights to have a common dimension with indices.")
    elif weights.ndim == 2:
        if weights.shape[0] != D:
            raise ValueError("Expected weights to have a common dimension with indices.")
    else:
        raise ValueError("Expected weights to be 1D or 2D.")

    (D, N) = weights.shape

    # Parse pathing
    if pathing is False:
        pathing = np.arange(terms.shape[0])
    if pathing is None:
        pathing = _term_pathing(terms)

    # Prepare the grids and canvas.
    (x_grid, y_grid) = _process_grid(grid)
    out = _parse_out(x_grid, out, stack=N)

    out.fill(0)
    nx0 = ny0 = 0
    if cp == np:
        xp = np
        monomial = np.ones_like(x_grid)
    else:
        xp = cp.get_array_module(x_grid)
        monomial = xp.ones_like(x_grid)

    # Force datatype for easier multiplication.
    weights = weights.astype(out.dtype)

    # Sum the result.
    for index in pathing:
        (nx, ny) = terms[index, :]

        if nx >= 0:                     # Usual case: monomial.
            # Reset if we're starting a new path.
            if nx - nx0 < 0 or ny - ny0 < 0:
                nx0 = ny0 = 0
                monomial.fill(1)

            # Traverse the path in +x or +y.
            for _ in range(nx - nx0):
                monomial *= x_grid
            for _ in range(ny - ny0):
                monomial *= y_grid

            # Update the current index.
            nx0, ny0 = nx, ny

            # We use a for loop here because the arrays will already be big (vectorization
            # overhead already amortized) and multiplying with zero or special indexing
            # can cost, esp. on GPU and scalar transfer is easier.
            for i in range(N):
                if weights[index, i] != 0:
                    out[i, ...] += weights[index, i] * monomial
        elif nx == -1 and ny == 0:      # Special case: vortex waveplate.
            if xp.iscomplexobj(x_grid):
                lg = xp.arctan2(xp.real(y_grid), xp.real(x_grid))
            else:
                lg = xp.arctan2(y_grid, x_grid)

            for i in range(N):
                if weights[index, i] > 0:
                    out[i, ...] += weights[index, i] * lg
        else:
            raise ValueError(f"Unrecognized terms {(nx, ny)} for index {index}.")

    return out


# Structured light.

def _determine_source_radius(grid, w=None):
    r"""
    Helper function to determine the assumed Gaussian source :math:`1/e` amplitude
    radius (:math:`1/e^2` power radius) for various
    structured light conversion functions. This is important because structured light
    conversions need knowledge of the size of the incident Gaussian beam.
    For example, see the ``w`` parameter in
    :meth:`~slmsuite.holography.toolbox.phase.laguerre_gaussian()`.

    Parameters
    ----------
    grid : (array_like, array_like) OR :class:`~slmsuite.hardware.slms.slm.SLM`
        Meshgrids of normalized :math:`\frac{x}{\lambda}` coordinates
        corresponding to SLM pixels, in ``(x_grid, y_grid)`` form.
        These are precalculated and stored in any :class:`~slmsuite.hardware.slms.slm.SLM`, so
        such a class can be passed instead of the grids directly.
    w : float OR None
        The radius of the phase pattern in normalized :math:`\frac{x}{\lambda}` units.
        To produce perfect structured beams, this radius is equal to the radius of
        the gaussian profile of the source (ideally not clipped by the SLM).
        If an SLM was passed as grid, retrieves the data from
        :attr:`slmsuite.hardware.slms.slm.SLM.source` and
        :meth:`slmsuite.hardware.slms.slm.SLM.fit_source_amplitude()`.
        If ``w`` is left as ``None``, ``w`` is set to a quarter of the smallest normalized screen dimension.

    Returns
    -------
    w : float
        Determined radius. In normalized units.
    """
    if w is not None:
        return w

    if hasattr(grid, "slm") and hasattr(grid, "cam"):
        grid = grid.slm
    if hasattr(grid, "get_source_radius"):
        return grid.get_source_radius()

    (x_grid, y_grid) = _process_grid(grid)
    return np.min([np.amax(x_grid), np.amax(y_grid)]) / 4


def laguerre_gaussian(grid, l, p=0, w=None):
    r"""
    Returns the phase farfield for a
    `Laguerre-Gaussian <https://en.wikipedia.org/wiki/Gaussian_beam#Laguerre-Gaussian_modes>`_
    beam. Uses the formalism described by
    `this paper <https://doi.org/10.1364/JOSAA.25.001642>`_.

    Note
    ~~~~
    Without radial order (``p = 0``), this function distills to a
    `vortex waveplate <https://en.wikipedia.org/wiki/Optical_vortex>`_
    of given azimuthal order ``l``.

    Tip
    ~~~
    This function is especially useful to hone and validate SLM alignment. Perfect alignment will
    result in concentric and uniform fringes for higher order beams. Focusing issues, aberration,
    or pointing misalignment will mitigate this.

    Parameters
    ----------
    grid : (array_like, array_like) OR :class:`~slmsuite.hardware.slms.slm.SLM`
        Meshgrids of normalized :math:`\frac{x}{\lambda}` coordinates
        corresponding to SLM pixels, in ``(x_grid, y_grid)`` form.
        These are precalculated and stored in any :class:`~slmsuite.hardware.slms.slm.SLM`, so
        such a class can be passed instead of the grids directly.
    l : int
        The azimuthal wavenumber, or orbital angular momentum. Can be negative.
    p : int
        The radial wavenumber. Should be non-negative.
    w : float OR None
        See :meth:`~slmsuite.holography.toolbox._determine_source_radius()`.

    Returns
    -------
    numpy.ndarray
        The phase for this function.
    """
    (x_grid, y_grid) = _process_grid(grid)

    w = _determine_source_radius(grid, w)

    theta_grid = np.arctan2(x_grid, y_grid)
    rr_grid = y_grid * y_grid + x_grid * x_grid

    canvas = 0

    if l != 0:
        canvas += l * theta_grid
    if p != 0:
        canvas += np.pi * np.heaviside(-special.genlaguerre(p, np.abs(l))(16 * rr_grid / w / w), 0)

    return canvas


def hermite_gaussian(grid, n, m, w=None):
    r"""
    Returns the phase farfield for a
    `Hermite-Gaussian <https://en.wikipedia.org/wiki/Gaussian_beam#Hermite-Gaussian_modes>`_
    beam. Uses the formalism described by `this paper <https://doi.org/10.1364/AO.54.008444>`_.

    Parameters
    ----------
    grid : (array_like, array_like) OR :class:`~slmsuite.hardware.slms.slm.SLM`
        Meshgrids of normalized :math:`\frac{x}{\lambda}` coordinates
        corresponding to SLM pixels, in ``(x_grid, y_grid)`` form.
        These are precalculated and stored in any :class:`~slmsuite.hardware.slms.slm.SLM`, so
        such a class can be passed instead of the grids directly.
    n, m : int
        The horizontal ``n`` and vertical ``m`` wavenumbers. ``n = m = 0`` yields a flat
        phase and a Gaussian beam.
    w : float
        See :meth:`~slmsuite.holography.toolbox._determine_source_radius()`.

    Returns
    -------
    numpy.ndarray
        The phase for this function.
    """
    (x_grid, y_grid) = _process_grid(grid)
    w = _determine_source_radius(grid, w)

    # factor = np.sqrt(2) / w
    factor = 4 / w

    # Generate the amplitude of a Hermite-Gaussian mode.
    phase = special.hermite(n)(factor * x_grid) * special.hermite(m)(factor * y_grid)

    # This is real, so the phase is just the sign of the mode. This produces a
    # checkerboard pattern. Probably could make this faster by bitflipping rows and columns.
    phase[phase < 0] = 0
    phase[phase > 0] = np.pi

    return phase


def ince_gaussian(grid, p, m, parity=1, ellipticity=1, w=None):
    r"""
    **(NotImplemented)** Returns the phase farfield for an
    `Ince-Gaussian <https://en.wikipedia.org/wiki/Gaussian_beam#Ince-Gaussian_modes>`_
    beam.
    `Consider <https://doi.org/10.1364/OL.29.000144>`_
    `using <https://doi.org/10.1364/AO.54.008444>`_
    `these <https://doi.org/10.3390/jimaging8050144>`_
    `references <https://en.wikipedia.org/wiki/Elliptic_coordinate_system>`_.

    Parameters
    ----------
    grid : (array_like, array_like) OR :class:`~slmsuite.hardware.slms.slm.SLM`
        Meshgrids of normalized :math:`\frac{x}{\lambda}` coordinates
        corresponding to SLM pixels, in ``(x_grid, y_grid)`` form.
        These are precalculated and stored in any :class:`~slmsuite.hardware.slms.slm.SLM`, so
        such a class can be passed instead of the grids directly.
    p : int
        Ince polynomial order.
    m : int
        Ince polynomial degree.
    parity : {1, -1, 0}
        Whether to produce an even (1), odd (-1), or helical (0) Ince polynomial. A helical
        polynomial is the linear combination of even and odd polynomials.

        .. math:: IG^h_{p,m} = IG^e_{p,m} + iIG^o_{p,m}

    ellipticity : float
        Ellipticity of the beam. The semifocal distance is equal to ``ellipticity * w``,
        where the foci are the points which define the elliptical coordinate system.
    w : float
        See :meth:`~slmsuite.holography.toolbox._determine_source_radius()`.

    Returns
    -------
    numpy.ndarray
        The phase for this function.
    """
    (x_grid, y_grid) = _process_grid(grid)
    w = _determine_source_radius(grid, w)

    if parity == 1:
        if not 0 <= m <= p:
            raise ValueError("{} is an invalid Ince polynomial.".format((p,m)))
    else:
        if not 1 <= m <= p:
            raise ValueError("{} is an invalid Ince polynomial.".format((p,m)))

    complex_grid = x_grid + 1j * y_grid

    factor = 1 / (w * np.sqrt(ellipticity / 2))

    elliptic_grid = np.arccosh(complex_grid * factor)

    raise NotImplementedError()


def matheui_gaussian(grid, r, q, w=None):
    """
    **(NotImplemented)** Returns the phase farfield for a
    `Matheui-Gaussian <https://doi.org/10.1364/AO.49.006903>`_ beam.

    Returns
    -------
    numpy.ndarray
        The phase for this function.
    """
    (x_grid, y_grid) = _process_grid(grid)
    w = _determine_source_radius(grid, w)

    raise NotImplementedError()


def airy(grid, f=(np.inf, np.inf)):
    """
    **(NotImplemented)** Returns the cubic phase farfield for an
    `Airy <http://dx.doi.org/10.1103/PhysRevLett.99.213901>`_ beam.

    Maybe also
    `parabolic <https://doi.org/10.1117/12.826193>`_,
    `symmetric Airy <http://dx.doi.org/10.1364/OL.39.002370>`_,
    or
    `circular symmetric Airy <https://doi.org/10.1016/j.optcom.2020.126190>`_
    beams.

    Returns
    -------
    numpy.ndarray
        The phase for this function.
    """
    (x_grid, y_grid) = _process_grid(grid)
    w = _determine_source_radius(grid, w)

    raise NotImplementedError()


"""
**(NotImplemented)** Hardware control for Holoeye SLMs.
This is a partial skeleton that can be completed if desired.
"""
import os
import warnings
from .slm import SLM

try:  # Load Holoeye's SDK module.
    from holoeye import slmdisplaysdk
except BaseException as e:  # Provide an informative error should something go wrong.
    warnings.warn(
        "Holoeye SDK failed to load.\n"
        "Original error: {}".format(e)
    )

class Holoeye(SLM):
    """
    Interfaces with Holoeye SLMs via the ``slmdisplaysdk``.

    Attributes
    ----------
    slm_lib : TODO
        Object handle for the Holoeye SLM.
    sdk_path : str
        Path of the Blink SDK folder.
    """

    def __init__(
        self,
        verbose=True,
        wav_um=1,
        pitch_um=(8,8),
        **kwargs
    ):
        r"""
        Initialize SLM and attributes.

        Parameters
        ----------
        wav_um : float
            Wavelength of operation in microns. Defaults to 1 um.
        pitch_um : (float, float)
            Pixel pitch in microns. Defaults to 8 micron square pixels.
        **kwargs
            See :meth:`.SLM.__init__` for permissible options.

        Note
        ~~~~
        These arguments, which ultimately are used to instantiate the :class:`.SLM` superclass,
        may be more accurately filled by calling the SLM's SDK functions.
        See the other implemented SLM subclasses for examples.
        """

        # Get the SLM.
        if verbose: print("Creating SLM instance...", end="")
        self.slm_lib = slmdisplaysdk.SLMInstance()  # ?
        # self.slm_lib = slmdisplaysdk.SLMDisplay() # ?

        # Check version somehow.
        if self.slm_lib.requiresVersion(3):
            if verbose: print("failure")
            raise RuntimeError("TODO")
        if verbose: print("success")

        # Open the SLM.
        if verbose: print("Opening SLM...", end="")
        error = self.slm_lib.open()

        if error != slmdisplaysdk.ErrorCode.NoError:
            if verbose: print("failure")
            self._handle_error(error)
        if verbose: print("success")

        # Other possibilities to consider:
        # - Setting the SLM's operating wavelength (wav_um).
        # - Updating the SLM's phase table if necessary, and/or setting the design
        #   wavelength (wav_design_um).
        # - Setting the SLM's default settle time (abstract class SLM uses
        #   settle_time_s=0.3 seconds). This is important for experimental feedback to
        #   allow the SLM to settle before viewing the result on a camera.
        # - Checking for and saving the SLM parameters (height, width, etc).

        # Instantiate the superclass
        super().__init__(
            (self.slm_lib.width(), self.slm_lib.height()),
            bitdepth=self.slm_lib.depth(),
            wav_um=wav_um,
            pitch_um=pitch_um,
            **kwargs
        )

        # Zero the display using the superclass `write()` function.
        self.write(None)

    def _handle_error(self, error):
        if error != slmdisplaysdk.ErrorCode.NoError:
            raise RuntimeError(self.slm_lib.errorString(error))

    @staticmethod
    def info(verbose=True):
        """
        Discovers all SLMs detected by an SDK.
        Useful for a user to identify the correct serial numbers / etc.

        Parameters
        ----------
        verbose : bool
            Whether to print the discovered information.

        Returns
        --------
        list of str
            List of serial numbers or identifiers.
        """
        raise NotImplementedError()
        serial_list = get_serial_list()     # TODO: Fill in proper function.
        return serial_list

    def _write_hw(self, phase):
        """
        Low-level hardware interface to write ``phase`` data onto the SLM.
        When the user calls the :meth:`.SLM.write` method of
        :class:`.SLM`, ``phase`` is error checked before calling
        :meth:`_write_hw()`. See :meth:`.SLM._write_hw` for further detail.
        """
        error = self.slm_lib.showData(phase, self.flags)
        self._handle_error(error)


"""
Light wrapper for `Instrumental cameras
<https://instrumental-lib.readthedocs.io/en/stable/install.html>`_.
The :mod:`instrumental` module must be installed ``pip install instrumental-lib``.

Note
~~~~
Color camera functionality is not currently implemented, and will lead to undefined behavior.
"""
from slmsuite.hardware.cameras.camera import Camera

try:
    import instrumental.drivers.cameras.Camera as InstrumentalCamera
except ImportError:
    print("instrumental-lib not installed. Install to use Instrumental cameras.")


class Instrumental(Camera):
    """
    Wrapped :mod:`instrumental` camera.

    Attributes
    ----------
    cam : instrumental.drivers.cameras.Camera
        Object to talk with the desired camera.
    exposure_s : float
        Instrumental doesn't save exposure. It sets the exposure at each
        :meth:`get_image`. This variable stores the desired exposure.
        Defaults to .001 (1 ms).
    """

    ### Initialization and termination ###

    def __init__(self, cam, **kwargs):
        """
        Initialize camera and attributes. Initial profile is ``"single"``.

        Parameters
        ----------
        cam : instrumental.drivers.cameras.Camera
            This class is just a wrapper for :mod:`instrumental`, so the user must pass a
            constructed :mod:`instrumental` camera.
        kwargs
            See :meth:`.Camera.__init__` for permissible options.

        Raises
        ------
        RuntimeError
           If the camera can not be reached.
        """
        assert isinstance(cam, InstrumentalCamera), "A subclass of instrumental.drivers.cameras.Camera must be passed as cam."
        self.cam = cam
        self.exposure_s = .001

        super().__init__(
            (self.cam.width, self.cam.height),
            bitdepth=8,         # Currently defaults to 8 because instrumental doesn't cache this. Update in the future, maybe.
            pitch_um=None,      # Currently unset because instrumental doesn't cache this. Update in the future, maybe.
            name=cam.__class__.__name__,
            **kwargs
        )

    def close(self):
        """
        See :meth:`.Camera.close`.
        """
        try:
            self.cam.close()
        except:
            raise RuntimeError("This instrumental camera does not support .close().")

    @staticmethod
    def info(verbose=True):
        """
        Method to load display information.

        Returns
        -------
        list
            An empty list.
        """
        raise RuntimeError(
            ".info() is not applicable to instrumental cameras, which are "
            "constructed outside this wrapper."
        )

    def reset(self):
        """
        See :meth:`.Camera.reset`.
        """
        raise RuntimeError("Instrumental cameras do not support reset.")

    def get_exposure(self):
        """
        Method to get the integration time in seconds.
        Used in :meth:`.autoexposure()`.

        Returns
        -------
        float
            Integration time in seconds.
        """
        return self.exposure_s

    def set_exposure(self, exposure_s):
        """
        Method to set the integration time in seconds.
        Used in :meth:`.autoexposure()`.

        Parameters
        ----------
        exposure_s : float
            The integration time in seconds.
        """
        self.exposure_s = exposure_s

    def set_woi(self, woi=None):
        """
        Method to narrow the imaging region to a 'window of interest'
        for faster framerates.

        Parameters
        ----------
        woi : list, None
            See :attr:`~slmsuite.hardware.cameras.camera.Camera.woi`.
            If ``None``, defaults to largest possible.

        Returns
        ----------
        woi : list
            :attr:`~slmsuite.hardware.cameras.camera.Camera.woi`.
        """
        raise NotImplementedError()

    def flush(self, timeout_s=1):
        """
        Method to cycle the image buffer (if any)
        such that all new :meth:`.get_image()`
        calls yield fresh frames.

        This is currently implemented by capturing five camera frames, and should be
        improved in the future.

        Parameters
        ----------
        timeout_s : float
            The time in seconds to wait for frames to catch up with triggers.
        """
        self.get_images(5)

    def _get_image_hw(self, timeout_s=1):
        """
        Method to pull an image from the camera and return.

        Parameters
        ----------
        timeout_s : float
            The time in seconds to wait for the frame to be fetched (currently unused).

        Returns
        -------
        numpy.ndarray
            Array of shape :attr:`~slmsuite.hardware.cameras.camera.Camera.shape`.
        """
        return self.cam.grab_image(timeouts=str(self.exposure_s) + "s", copy=True)


"""
**(NotImplemented)** Hardware control for FLIR cameras via the :mod:`PySpin` interface to the Spinnaker SDK.
Install Spinnaker by following the
`provided instructions <https://www.flir.com/products/spinnaker-sdk/>`_.
Inspired by `this implementation <https://github.com/klecknerlab/simple_pyspin/>`_.

Warning
~~~~~~~~
Implementation unfinished and untested. Consider using ``simple_pyspin`` as a dependency instead.
"""
import warnings

from .camera import Camera

try:
    import PySpin
except ImportError:
    warnings.warn("PySpin not installed. Install to use FLIR cameras.")


class FLIR(Camera):
    """
    FLIR camera.

    Attributes
    ----------
    sdk : PySpin.System
        Spinnaker SDK.
    cam : PySpin.Camera
        Object to talk with the desired camera.
    """

    sdk = None

    ### Initialization and termination ###

    def __init__(self, serial="", pitch_um=None, verbose=True, **kwargs):
        """
        Initialize camera and attributes.

        Parameters
        ----------
        serial : str
            Serial number of the camera to open.
        pitch_um : (float, float) OR None
            Fill in extra information about the pixel pitch in ``(dx_um, dy_um)`` form
            to use additional calibrations.
        verbose : bool
            Whether or not to log.
        **kwargs
            See :meth:`.Camera.__init__` for permissible options.
        """
        if FLIR.sdk is None:
            if verbose:
                print("PySpin initializing... ", end="")
            FLIR.sdk = PySpin.System.get_instance()
            if verbose:
                print("success")

        if verbose:
            print("Looking for cameras... ", end="")
        camera_list = PySpin.sdk.GetCameras()
        if verbose:
            print("success")

        if verbose:
            print("Camera initializing... ", end="")
        if serial == "":
            self.cam = camera_list.GetByIndex(0)
        else:
            self.cam = camera_list.GetBySerial(serial)
        self.cam.Init()
        if verbose:
            print("success")

        super().__init__(
            (self.cam.SensorWidth.get(), self.cam.SensorHeight.get()),
            bitdepth=int(self.cam.PixelSize.get()),
            pitch_um=pitch_um,
            name=serial,
            **kwargs
        )

        raise NotImplementedError()

    def close(self, close_sdk=True):
        """See :meth:`.Camera.close`."""
        try:
            self.cam.EndAcquisition()
        except BaseException:
            pass
        del self.cam

    ### Property Configuration ###

    def get_exposure(self):
        """See :meth:`.Camera.get_exposure`."""
        return

    def set_exposure(self, exposure_s):
        """See :meth:`.Camera.set_exposure`."""
        return

    def set_woi(self, window=None):
        """See :meth:`.Camera.set_woi`."""
        return

    def _get_image_hw(self, blocking=True):
        """
        See :meth:`.Camera._get_image_hw`.

        Parameters
        ----------
        blocking : bool
            Whether to wait for the camera to return a frame, blocking other acquisition.
        """
        frame = self.cam.GetNextImage(
            PySpin.EVENT_TIMEOUT_INFINITE if blocking else PySpin.EVENT_TIMEOUT_NONE
        )

        return frame.GetNDArray()



"""
Hardware control for AlliedVision cameras via the :mod:`vimba` interface.
Install :mod:`vimba` by following the
`provided instructions <https://github.com/alliedvision/VimbaPython>`_.
Be sure to include the ``numpy-export`` flag in the ``pip install`` command,
as the :class:`AlliedVision` class makes use of these features. See especially the
`vimba python manual <https://github.com/alliedvision/VimbaPython/blob/master/Documentation/Vimba%20Python%20Manual.pdf>`_
for reference.

Note
~~~~
Color camera functionality is not currently implemented, and will lead to undefined behavior.
"""

import time
import numpy as np
import warnings

from slmsuite.hardware.cameras.camera import Camera

try:
    import vimba
except ImportError:
    warnings.warn("vimba not installed. Install to use AlliedVision cameras.")


class AlliedVision(Camera):
    r"""
    AlliedVision camera.

    Attributes
    ----------
    sdk : vimba.Vimba
        AlliedVision SDK. Shared among instances of :class:`AlliedVision`.
    cam : vimba.Camera
        Object to talk with the desired camera.

    Caution
    ~~~~~~~~
    The AlliedVision SDK :mod:`vimba` includes protections to maintain camera connectivity:
    specifically, the SDK :class:`vimba.Vimba` and cameras :class:`vimba.Camera` are designed
    to be used in concert with ``with`` statements. Unfortunately, this does not mesh with the
    architecture of :mod:`slmsuite`, where notebook-style operation is desired.
    Using ``with`` statements inside :class:`.AlliedVision` methods is likewise not an option,
    as the methods to :meth:`__enter__()` and :meth:`__exit__()` the ``with`` are time-consuming
    due to calls to :meth:`_open()` and :meth:`_close()` the objects, to the point of
    :math:`\mathcal{O}(\text{s})` overhead. :class:`.AlliedVision` disables these protections by
    calling :meth:`__enter__()` and :meth:`__exit__()` directly during :meth:`.__init__()` and
    :meth:`.close()`, instead of inside ``with`` statements.
    """

    sdk = None

    def __init__(self, serial="", pitch_um=None, verbose=True, **kwargs):
        """
        Initialize camera and attributes.

        Parameters
        ----------
        serial : str
            Serial number of the camera to open. If empty, defaults to the first camera in the list
            returned by :meth:`vimba.get_all_cameras()`.
        pitch_um : (float, float) OR None
            Fill in extra information about the pixel pitch in ``(dx_um, dy_um)`` form
            to use additional calibrations.
        verbose : bool
            Whether or not to print extra information.
        **kwargs
            See :meth:`.Camera.__init__` for permissible options.
        """
        if AlliedVision.sdk is None:
            if verbose:
                print("vimba initializing... ", end="")
            AlliedVision.sdk = vimba.Vimba.get_instance()
            AlliedVision.sdk.__enter__()
            if verbose:
                print("success")

        if verbose:
            print("Looking for cameras... ", end="")
        camera_list = AlliedVision.sdk.get_all_cameras()
        if verbose:
            print("success")

        serial_list = [cam.get_serial() for cam in camera_list]
        if serial == "":
            if len(camera_list) == 0:
                raise RuntimeError("No cameras found by vimba.")
            if len(camera_list) > 1 and verbose:
                print(f"No serial given... Choosing first of {serial_list}")

            self.cam = camera_list[0]
            serial = self.cam.get_serial()
        else:
            if serial in serial_list:
                self.cam = camera_list[serial_list.index(serial)]
            else:
                raise RuntimeError(
                    f"Serial {serial} not found by vimba. Available: {serial_list}"
                )

        if verbose:
            print(f"vimba sn '{serial}' initializing... ", end="")
        self.cam.__enter__()
        if verbose:
            print("success")

        super().__init__(
            (self.cam.SensorWidth.get(), self.cam.SensorHeight.get()),
            bitdepth=int(self.cam.PixelSize.get()),
            pitch_um=pitch_um,
            name=serial,
            **kwargs
        )

        try:
            self.cam.BinningHorizontal.set(1)
            self.cam.BinningVertical.set(1)
        except:
            pass    # Some cameras do not have the option to set binning.

        self.cam.GainAuto.set("Off")

        self.cam.ExposureAuto.set("Off")
        self.cam.ExposureMode.set("Timed")

        self.cam.AcquisitionMode.set("SingleFrame")

        # Future: triggered instead of SingleFrame.
        self.cam.TriggerSelector.set("AcquisitionStart")
        self.cam.TriggerMode.set("Off")
        self.cam.TriggerActivation.set("RisingEdge")
        self.cam.TriggerSource.set("Software")

    def close(self, close_sdk=True):
        """
        See :meth:`.Camera.close`

        Parameters
        ----------
        close_sdk : bool
            Whether or not to close the :mod:`vimba` instance.
        """
        self.cam.__exit__(None, None, None)

        if close_sdk:
            self.close_sdk()

        del self.cam

    @staticmethod
    def info(verbose=True):
        """
        Discovers all Thorlabs scientific cameras.

        Parameters
        ----------
        verbose : bool
            Whether to print the discovered information.

        Returns
        --------
        list of str
            List of AlliedVision serial numbers.
        """
        if AlliedVision.sdk is None:
            AlliedVision.sdk = vimba.Vimba.get_instance()
            AlliedVision.sdk.__enter__()
            close_sdk = True
        else:
            close_sdk = False

        camera_list = AlliedVision.sdk.get_all_cameras()
        serial_list = [cam.get_serial() for cam in camera_list]

        if verbose:
            print("AlliedVision serials:")
            for serial in serial_list:
                print(f"'{serial}'")

        if close_sdk:
            AlliedVision.close_sdk()

        return serial_list

    @classmethod
    def close_sdk(cls):
        """
        Close the :mod:`vimba` instance.
        """
        if cls.sdk is not None:
            cls.sdk.__exit__(None, None, None)
            cls.sdk = None

    ### Property Configuration ###

    def get_properties(self, properties=None):
        """
        Print the list of camera properties.

        Parameters
        ----------
        properties : dict or None
            The target camera's property dictionary. If ``None``, the property
            dictionary is fetched from the camera associated with the calling instance.
        """
        if properties is None:
            properties = self.cam.__dict__.keys()

        for key in properties:
            prop = self.cam.__dict__[key]
            try:
                print(prop.get_name(), end="\t")
            except BaseException as e:
                print(f"Error accessing property dictionary, '{key}':{e}")
                continue

            try:
                print(prop.get(), end="\t")
            except:
                pass

            try:
                print(prop.get_unit(), end="\t")
            except:
                pass

            try:
                print(prop.get_description(), end="\n")
            except:
                print("")

    def set_adc_bitdepth(self, bitdepth):
        """
        Set the digitization bitdepth.

        Parameters
        ----------
        bitdepth : int
            Desired digitization bitdepth.
        """
        bitdepth = int(bitdepth)

        for entry in self.cam.SensorBitDepth.get_all_entries():
            value = entry.as_tuple()  # (name : str, value : int)
            if str(bitdepth) in value[0]:
                self.cam.SensorBitDepth.set(value[1])
                break
            raise RuntimeError(f"ADC bitdepth {bitdepth} not found.")

    def get_adc_bitdepth(self):
        """
        Get the digitization bitdepth.

        Returns
        -------
        int
            The digitization bitdepth.
        """
        value = str(self.cam.SensorBitDepth.get())
        bitdepth = int("".join(char for char in value if char.isdigit()))
        return bitdepth

    def get_exposure(self):
        """See :meth:`.Camera.get_exposure`."""
        return float(self.cam.ExposureTime.get()) / 1e6

    def set_exposure(self, exposure_s):
        """See :meth:`.Camera.set_exposure`."""
        self.cam.ExposureTime.set(float(exposure_s * 1e6))

    def set_woi(self, woi=None):
        """See :meth:`.Camera.set_woi`."""
        return

    def _get_image_hw(self, timeout_s=5):
        """See :meth:`.Camera._get_image_hw`."""
        t = time.time()

        # Convert timeout_s to ms
        frame = self.cam.get_frame(timeout_ms=int(1e3 * timeout_s))
        frame = frame.as_numpy_ndarray()

        # We have noticed that sometimes the camera gets into a state where
        # it returns a frame of all zeros apart from one pixel with value of 31.
        # This method is admittedly a hack to try getting a frame a few more times.
        # We welcome contributions to fix this.
        while np.sum(frame) == np.amax(frame) == 31 and time.time() - t < timeout_s:
            frame = self.cam.get_frame(timeout_ms=int(1e3 * timeout_s))
            frame = frame.as_numpy_ndarray()

        return np.squeeze(frame)

    def flush(self, timeout_s=1):
        """See :meth:`.Camera.flush`."""
        for _ in range(2):
            self._get_image_hw_tolerant()

    def reset(self):
        """See :meth:`.Camera.reset`."""
        raise NotImplementedError()
